"""
Topological fingerprints for macromolecular structures.
"""
import numpy as np
import logging
import itertools
from deepchem.utils.hash_utils import hash_ecfp
from deepchem.feat import ComplexFeaturizer
from deepchem.utils.rdkit_utils import load_complex
from deepchem.utils.hash_utils import vectorize
from deepchem.utils.voxel_utils import voxelize
from deepchem.utils.voxel_utils import convert_atom_to_voxel
from deepchem.utils.rdkit_utils import compute_all_ecfp
from deepchem.utils.rdkit_utils import compute_contact_centroid
from deepchem.utils.rdkit_utils import MoleculeLoadException
from deepchem.utils.geometry_utils import compute_pairwise_distances
from deepchem.utils.geometry_utils import subtract_centroid

from typing import Optional, Tuple, Dict

logger = logging.getLogger(__name__)


def featurize_contacts_ecfp(
        frag1: Tuple,
        frag2: Tuple,
        pairwise_distances: Optional[np.ndarray] = None,
        cutoff: float = 4.5,
        ecfp_degree: int = 2) -> Tuple[Dict[int, str], Dict[int, str]]:
    """Computes ECFP dicts for pairwise interaction between two molecular fragments.

    Parameters
    ----------
    frag1: Tuple
        A tuple of (coords, mol) returned by `load_molecule`.
    frag2: Tuple
        A tuple of (coords, mol) returned by `load_molecule`.
    pairwise_distances: np.ndarray
        Array of pairwise fragment-fragment distances (Angstroms)
    cutoff: float
        Cutoff distance for contact consideration
    ecfp_degree: int
        ECFP radius

    Returns
    -------
    Tuple of dictionaries of ECFP contact fragments
    """
    if pairwise_distances is None:
        pairwise_distances = compute_pairwise_distances(frag1[0], frag2[0])
    # contacts is of form (x_coords, y_coords), a tuple of 2 lists
    contacts = np.nonzero((pairwise_distances < cutoff))
    # contacts[0] is the x_coords, that is the frag1 atoms that have
    # nonzero contact.
    frag1_atoms = set([int(c) for c in contacts[0].tolist()])
    # contacts[1] is the y_coords, the frag2 atoms with nonzero contacts
    frag2_atoms = set([int(c) for c in contacts[1].tolist()])

    frag1_ecfp_dict = compute_all_ecfp(frag1[1],
                                       indices=frag1_atoms,
                                       degree=ecfp_degree)
    frag2_ecfp_dict = compute_all_ecfp(frag2[1],
                                       indices=frag2_atoms,
                                       degree=ecfp_degree)

    return (frag1_ecfp_dict, frag2_ecfp_dict)


class ContactCircularFingerprint(ComplexFeaturizer):
    """Compute (Morgan) fingerprints near contact points of macromolecular complexes.

    Given a macromolecular complex made up of multiple
    constituent molecules, first compute the contact points where
    atoms from different molecules come close to one another. For
    atoms within "contact regions," compute radial "ECFP"
    fragments which are sub-molecules centered at atoms in the
    contact region.

    For a macromolecular complex, returns a vector of shape
    `(2*size,)`
    """

    def __init__(self, cutoff: float = 4.5, radius: int = 2, size: int = 8):
        """
        Parameters
        ----------
        cutoff: float (default 4.5)
            Distance cutoff in angstroms for molecules in complex.
        radius: int, optional (default 2)
            Fingerprint radius.
        size: int, optional (default 8)
            Length of generated bit vector.
        """
        self.cutoff = cutoff
        self.radius = radius
        self.size = size

    def _featurize(self, datapoint, **kwargs):
        """
        Compute featurization for a molecular complex

        Parameters
        ----------
        datapoint: Tuple[str, str]
            Filenames for molecule and protein.
        """
        if 'complex' in kwargs:
            datapoint = kwargs.get("complex")
            raise DeprecationWarning(
                'Complex is being phased out as a parameter, please pass "datapoint" instead.'
            )

        try:
            fragments = load_complex(datapoint, add_hydrogens=False)

        except MoleculeLoadException:
            logger.warning(
                "This molecule cannot be loaded by Rdkit. Returning None")
            return None
        pairwise_features = []
        # We compute pairwise contact fingerprints
        for (frag1, frag2) in itertools.combinations(fragments, 2):
            # Get coordinates
            distances = compute_pairwise_distances(frag1[0], frag2[0])
            vector = [
                vectorize(hash_ecfp, feature_dict=ecfp_dict, size=self.size) for
                ecfp_dict in featurize_contacts_ecfp(frag1,
                                                     frag2,
                                                     distances,
                                                     cutoff=self.cutoff,
                                                     ecfp_degree=self.radius)
            ]
            pairwise_features += vector

        pairwise_features = np.concatenate(pairwise_features)
        return pairwise_features


class ContactCircularVoxelizer(ComplexFeaturizer):
    """Computes ECFP fingerprints on a voxel grid.

    Given a macromolecular complex made up of multiple
    constituent molecules, first compute the contact points where
    atoms from different molecules come close to one another. For
    atoms within "contact regions," compute radial "ECFP"
    fragments which are sub-molecules centered at atoms in the
    contact region. Localize these ECFP fingeprints at the voxel
    in which they originated.

    Featurizes a macromolecular complex into a tensor of shape
    `(voxels_per_edge, voxels_per_edge, voxels_per_edge, size)` where
    `voxels_per_edge = int(box_width/voxel_width)`. If `flatten==True`,
    then returns a flattened version of this tensor of length
    `size*voxels_per_edge**3`
    """

    def __init__(self,
                 cutoff: float = 4.5,
                 radius: int = 2,
                 size: int = 8,
                 box_width: float = 16.0,
                 voxel_width: float = 1.0,
                 flatten: bool = False):
        """
        Parameters
        ----------
        cutoff: float (default 4.5)
            Distance cutoff in angstroms for molecules in complex.
        radius : int, optional (default 2)
            Fingerprint radius.
        size : int, optional (default 8)
            Length of generated bit vector.
        box_width: float, optional (default 16.0)
            Size of a box in which voxel features are calculated. Box
            is centered on a ligand centroid.
        voxel_width: float, optional (default 1.0)
            Size of a 3D voxel in a grid.
        flatten: bool, optional (default False)
            If True, then returns a flat feature vector rather than voxel grid. This
            feature vector is constructed by flattening the usual voxel grid.
        """
        self.cutoff = cutoff
        self.radius = radius
        self.size = size
        self.box_width = box_width
        self.voxel_width = voxel_width
        self.voxels_per_edge = int(self.box_width / self.voxel_width)
        self.flatten = flatten

    def _featurize(self, complex):
        """
        Compute featurization for a molecular complex

        Parameters
        ----------
        complex: Tuple[str, str]
            Filenames for molecule and protein.
        """
        try:
            fragments = load_complex(complex, add_hydrogens=False)

        except MoleculeLoadException:
            logger.warning(
                "This molecule cannot be loaded by Rdkit. Returning None")
            return None
        pairwise_features = []
        # We compute pairwise contact fingerprints
        centroid = compute_contact_centroid(fragments, cutoff=self.cutoff)
        for (frag1, frag2) in itertools.combinations(fragments, 2):
            distances = compute_pairwise_distances(frag1[0], frag2[0])
            frag1_xyz = subtract_centroid(frag1[0], centroid)
            frag2_xyz = subtract_centroid(frag2[0], centroid)
            xyzs = [frag1_xyz, frag2_xyz]
            pairwise_features.append(
                sum([
                    voxelize(convert_atom_to_voxel,
                             xyz,
                             self.box_width,
                             self.voxel_width,
                             hash_function=hash_ecfp,
                             feature_dict=ecfp_dict,
                             nb_channel=self.size)
                    for xyz, ecfp_dict in zip(
                        xyzs,
                        featurize_contacts_ecfp(frag1,
                                                frag2,
                                                distances,
                                                cutoff=self.cutoff,
                                                ecfp_degree=self.radius))
                ]))
        if self.flatten:
            return np.concatenate(
                [features.flatten() for features in pairwise_features])
        else:
            # Features are of shape (voxels_per_edge, voxels_per_edge,
            # voxels_per_edge, num_feat) so we should concatenate on the last
            # axis.
            return np.concatenate(pairwise_features, axis=-1)


def compute_all_sybyl(mol, indices=None):
    """Computes Sybyl atom types for atoms in molecule."""
    raise NotImplementedError("This function is not implemented yet")


def featurize_binding_pocket_sybyl(protein_xyz,
                                   protein,
                                   ligand_xyz,
                                   ligand,
                                   pairwise_distances=None,
                                   cutoff=7.0):
    """Computes Sybyl dicts for ligand and binding pocket of the protein.

    Parameters
    ----------
    protein_xyz: np.ndarray
        Of shape (N_protein_atoms, 3)
    protein: Rdkit Molecule
        Contains more metadata.
    ligand_xyz: np.ndarray
        Of shape (N_ligand_atoms, 3)
    ligand: Rdkit Molecule
        Contains more metadata
    pairwise_distances: np.ndarray
        Array of pairwise protein-ligand distances (Angstroms)
    cutoff: float
        Cutoff distance for contact consideration.
    """

    if pairwise_distances is None:
        pairwise_distances = compute_pairwise_distances(protein_xyz, ligand_xyz)
    contacts = np.nonzero((pairwise_distances < cutoff))
    protein_atoms = set([int(c) for c in contacts[0].tolist()])

    protein_sybyl_dict = compute_all_sybyl(protein, indices=protein_atoms)
    ligand_sybyl_dict = compute_all_sybyl(ligand)
    return (protein_sybyl_dict, ligand_sybyl_dict)



import torch
import torch.nn as nn
from deepchem.utils.typing import OneOrMany
from deepchem.models.torch_models.layers import HighwayLayer, DTNNEmbedding
import numpy as np
from deepchem.models.torch_models.torch_model import TorchModel
from deepchem.models.losses import L2Loss, SoftmaxCrossEntropy
from typing import List, Tuple, Iterable, Union, Any, Dict
from deepchem.data import Dataset
from deepchem.metrics import to_one_hot
import copy
import sys

default_dict = {
    '#': 1,
    '(': 2,
    ')': 3,
    '+': 4,
    '-': 5,
    '/': 6,
    '1': 7,
    '2': 8,
    '3': 9,
    '4': 10,
    '5': 11,
    '6': 12,
    '7': 13,
    '8': 14,
    '=': 15,
    'C': 16,
    'F': 17,
    'H': 18,
    'I': 19,
    'N': 20,
    'O': 21,
    'P': 22,
    'S': 23,
    '[': 24,
    '\\': 25,
    ']': 26,
    '_': 27,
    'c': 28,
    'Cl': 29,
    'Br': 30,
    'n': 31,
    'o': 32,
    's': 33
}


class TextCNN(nn.Module):
    """
    A 1D convolutional neural network for both classification and regression tasks.

    Reimplementation of the discriminator module in ORGAN [1] .
    Originated from [2].

    The model converts the input smile strings to an embedding vector, the vector
    is convolved and pooled through a series of convolutional filters which are concatnated
    and later passed through a simple dense layer. The resulting vector goes through a Highway
    layer [3] which finally as per the nature of the task is passed through a dense layer.

    References
    ----------
    .. [1]  Guimaraes, Gabriel Lima, et al. "Objective-reinforced generative adversarial networks (ORGAN) for sequence generation models." arXiv preprint arXiv:1705.10843 (2017).
    .. [2] Kim, Yoon. "Convolutional neural networks for sentence classification." arXiv preprint arXiv:1408.5882 (2014).
    .. [3] Srivastava et al., "Training Very Deep Networks".https://arxiv.org/abs/1507.06228

    Examples
    --------
    >>> from deepchem.models.torch_models.text_cnn import default_dict, TextCNN
    >>> import torch
    >>> batch_size = 1
    >>> input_tensor = torch.randint(34, (batch_size, 64))
    >>> cls_model = TextCNN(1, default_dict, 1, mode="classification")
    >>> reg_model = TextCNN(1, default_dict, 1, mode="regression")
    >>> cls_output = cls_model.forward(input_tensor)
    >>> reg_output = reg_model.forward(input_tensor)
    >>> assert len(cls_output) == 2
    >>> assert len(reg_output) == 1
    """

    def __init__(self,
                 n_tasks: int,
                 char_dict: Dict[str, int],
                 seq_length: int,
                 n_embedding: int = 75,
                 kernel_sizes: List[int] = [
                     1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 15, 20
                 ],
                 num_filters: List[int] = [
                     100, 200, 200, 200, 200, 100, 100, 100, 100, 100, 160, 160
                 ],
                 dropout: float = 0.25,
                 mode: str = "classification") -> None:
        """
        Parameters
        ----------
        n_tasks: int
            Number of tasks
        char_dict: dict
            Mapping from characters in smiles to integers
        seq_length: int
            Length of sequences(after padding)
        n_embedding: int, optional
            Length of embedding vector
        kernel_sizes: list of int, optional
            Properties of filters used in the conv net
        num_filters: list of int, optional
            Properties of filters used in the conv net
        dropout: float, optional
            Dropout rate
        mode: str
            Either "classification" or "regression" for type of model.
        """

        super(TextCNN, self).__init__()

        self.n_tasks = n_tasks
        self.char_dict = char_dict
        self.seq_length = max(seq_length, max(kernel_sizes))
        self.n_embedding = n_embedding
        self.kernel_sizes = kernel_sizes
        self.num_filters = num_filters
        self.dropout = dropout
        self.mode = mode

        self.conv_layers = nn.ModuleList()
        self.embedding_layer = DTNNEmbedding(
            n_embedding=self.n_embedding,
            periodic_table_length=len(self.char_dict.keys()) + 1)
        self.dropout_layer = nn.Dropout1d(p=self.dropout)
        for filter_size, num_filter in zip(self.kernel_sizes, self.num_filters):
            self.conv_layers.append(
                nn.Conv1d(in_channels=self.n_embedding,
                          out_channels=num_filter,
                          kernel_size=filter_size,
                          padding=0,
                          dtype=torch.float32))
        concat_emb_dim = sum(num_filters)
        self.linear1 = nn.Linear(in_features=concat_emb_dim, out_features=200)
        if (self.mode == "classification"):
            self.linear2 = nn.Linear(in_features=200,
                                     out_features=self.n_tasks * 2)
        else:
            self.linear2 = nn.Linear(in_features=200,
                                     out_features=self.n_tasks * 1)
        self.relu = nn.ReLU()
        self.softmax = nn.Softmax(dim=2)
        self.highway = HighwayLayer(200)

    def forward(self, input: OneOrMany[torch.Tensor]) -> List[Any]:
        """
        Parameters
        ----------
        input: torch.Tensor
            Input Tensor
        Returns
        -------
        torch.Tensor
            Output as per use case : regression/classification
        """
        input_emb = self.embedding_layer(input)
        input_emb = input_emb.permute(0, 2, 1)

        conv_outputs = []
        for i, conv_layer in enumerate(self.conv_layers):
            x = conv_layer(input_emb)
            x, _ = torch.max(x, dim=2)
            conv_outputs.append(x)
            if (i == 0):
                concat_output = x
            else:
                concat_output = torch.cat((concat_output, x), dim=1)

        x = self.relu(self.linear1(self.dropout_layer(concat_output)))
        x = self.highway(x)

        if self.mode == "classification":
            logits = self.linear2(x)
            logits = logits.view(-1, self.n_tasks, 2)
            output = self.softmax(logits)
            outputs = [output, logits]
        else:
            output = self.linear2(x)
            output = output.view(-1, self.n_tasks, 1)
            outputs = [output]
        return outputs


class TextCNNModel(TorchModel):
    """
   A 1D convolutional neural network to work on smiles strings for both
   classification and regression tasks.


   Reimplementation of the discriminator module in ORGAN [1] .
   Originated from [2].


   The model converts the input smile strings to an embedding vector, the vector
   is convolved and pooled through a series of convolutional filters which are concatnated
   and later passed through a simple dense layer. The resulting vector goes through a Highway
   layer [3] which finally as per the nature of the task is passed through a dense layer.


   References
   ----------
   .. [1]  Guimaraes, Gabriel Lima, et al. "Objective-reinforced generative adversarial networks (ORGAN) for sequence generation models." arXiv preprint arXiv:1705.10843 (2017).
   .. [2] Kim, Yoon. "Convolutional neural networks for sentence classification." arXiv preprint arXiv:1408.5882 (2014).
   .. [3] Srivastava et al., "Training Very Deep Networks".https://arxiv.org/abs/1507.06228


   Examples
   --------
   >>> import os
   >>> from deepchem.models.torch_models import TextCNNModel
   >>> from deepchem.models.torch_models.text_cnn import default_dict
   >>> n_tasks = 1
   >>> seq_len = 250
   >>> model = TextCNNModel(n_tasks, default_dict, seq_len)
   """

    def __init__(self,
                 n_tasks: int,
                 char_dict: Dict[str, int],
                 seq_length: int,
                 n_embedding: int = 75,
                 kernel_sizes: List[int] = [
                     1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 15, 20
                 ],
                 num_filters: List[int] = [
                     100, 200, 200, 200, 200, 100, 100, 100, 100, 100, 160, 160
                 ],
                 dropout: float = 0.25,
                 mode: str = "classification",
                 **kwargs) -> None:
        """
       Parameters
       ----------
       n_tasks: int
           Number of tasks
       char_dict: dict
           Mapping from characters in smiles to integers
       seq_length: int
           Length of sequences(after padding)
       n_embedding: int, optional
           Length of embedding vector
       filter_sizes: list of int, optional
           Properties of filters used in the conv net
       num_filters: list of int, optional
           Properties of filters used in the conv net
       dropout: float, optional
           Dropout rate
       mode: str
           Either "classification" or "regression" for type of model.
       """

        self.n_tasks = n_tasks
        self.char_dict = char_dict
        self.seq_length = max(seq_length, max(kernel_sizes))
        self.n_embedding = n_embedding
        self.kernel_sizes = kernel_sizes
        self.num_filters = num_filters
        self.dropout = dropout
        self.mode = mode

        self.model = TextCNN(
            n_tasks=n_tasks,
            char_dict=char_dict,
            seq_length=seq_length,
            n_embedding=n_embedding,
            kernel_sizes=kernel_sizes,
            num_filters=num_filters,
            dropout=dropout,
            mode=mode,
        )
        loss: Union[SoftmaxCrossEntropy, L2Loss]
        if self.mode == "classification":

            loss = SoftmaxCrossEntropy()
            output_types = ['prediction', 'loss']
        else:
            loss = L2Loss()
            output_types = ['prediction']

        super(TextCNNModel, self).__init__(self.model,
                                           loss=loss,
                                           output_types=output_types,
                                           **kwargs)

    # Below functions were taken from DeepChem TextCNN tensorflow implementation

    def default_generator(
            self,
            dataset: Dataset,
            epochs: int = 1,
            mode: str = 'fit',
            deterministic: bool = True,
            pad_batches: bool = True) -> Iterable[Tuple[List, List, List]]:
        """
        Transfer smiles strings to fixed length integer vectors

        Parameters
        ----------
        dataset: `dc.data.Dataset`
            Dataset to convert
        epochs: int, optional (Default 1)
            Number of times to walk over `dataset`
        mode: str, optional (Default 'fit')
            Ignored in this implementation.
        deterministic: bool, optional (Default True)
            Whether the dataset should be walked in a deterministic fashion
        pad_batches: bool, optional (Default True)
            If true, each returned batch will have size `self.batch_size`.

        Returns
        -------
        Iterator which walks over the batches
        """
        for epoch in range(epochs):
            for (X_b, y_b, w_b,
                 ids_b) in dataset.iterbatches(batch_size=self.batch_size,
                                               deterministic=deterministic,
                                               pad_batches=pad_batches):
                if y_b is not None:
                    if self.mode == 'classification':
                        y_b = to_one_hot(y_b.flatten(),
                                         2).reshape(-1, self.n_tasks, 2)
                # Transform SMILES sequence to integers

                X_b = self.smiles_to_seq_batch(ids_b)
                yield ([X_b], [y_b], [w_b])

    @staticmethod
    def build_char_dict(dataset: Dataset,
                        default_dict: Dict[str, int] = default_dict):
        """
        Collect all unique characters(in smiles) from the dataset.
        This method should be called before defining the model to build appropriate char_dict

        Parameters
        ----------
        dataset: Dataset
           Dataset for which char_dict is built for
        default_dict: dict, optional
           Mapping from characters in smiles to integers, optional

        Returns
        -------
        out_dict: dict
            A dictionary containing mapping between unique characters in the dataset to integers
        seq_length: int
            The maximum sequence length of smile strings found in the dataset multiplied by 1.2
        """
        X = dataset.ids
        # Maximum length is expanded to allow length variation during train and inference
        seq_length = int(max([len(smile) for smile in X]) * 1.2)
        # '_' served as delimiter and padding
        all_smiles = '_'.join(X)
        tot_len = len(all_smiles)
        # Initialize common characters as keys
        keys = list(default_dict.keys())
        out_dict = copy.deepcopy(default_dict)
        current_key_val = len(keys) + 1
        # Include space to avoid extra keys
        keys.extend([' '])
        extra_keys = []
        i = 0
        while i < tot_len:
            # For 'Cl', 'Br', etc.
            if all_smiles[i:i + 2] in keys:
                i = i + 2
            elif all_smiles[i:i + 1] in keys:
                i = i + 1
            else:
                # Character not recognized, add to extra_keys
                extra_keys.append(all_smiles[i])
                keys.append(all_smiles[i])
                i = i + 1
        # Add all extra_keys to char_dict
        for extra_key in extra_keys:
            out_dict[extra_key] = current_key_val
            current_key_val += 1
        return out_dict, seq_length

    def smiles_to_seq(self, smiles: str):
        """
        Tokenize characters in smiles to integers

        Parameters
        ----------
        smiles: str
           A smile string

        Returns
        -------
        array: np.ndarray
            An array of integers representing the tokenized sequence of characters.
        """
        smiles_len = len(smiles)
        seq = [0]
        keys = self.char_dict.keys()
        i = 0
        while i < smiles_len:
            # Skip all spaces
            if smiles[i:i + 1] == ' ':
                i = i + 1
            # For 'Cl', 'Br', etc.
            elif smiles[i:i + 2] in keys:
                seq.append(self.char_dict[smiles[i:i + 2]])
                i = i + 2
            elif smiles[i:i + 1] in keys:
                seq.append(self.char_dict[smiles[i:i + 1]])
                i = i + 1
            else:
                raise ValueError('character not found in dict')
        for i in range(self.seq_length - len(seq)):
            # Padding with '_'
            seq.append(self.char_dict['_'])
        return np.array(seq, dtype=np.int32)

    @staticmethod
    def convert_bytes_to_char(s: bytes) -> str:
        """
        Convert bytes to string.

        Parameters
        ----------
        s: bytes
            Bytes to be converted to string.

        Returns
        -------
        str
            String representation of the bytes.
        """
        out_str = ''.join(chr(b) for b in s)
        return out_str

    def smiles_to_seq_batch(
            self, ids_b: Union[List[Union[bytes, str]],
                               np.ndarray]) -> np.ndarray:
        """
        Converts SMILES strings to np.array sequence.

        Parameters
        ----------
        ids_b: Union[List[Union[bytes, str]], np.ndarray]
            A list of SMILES strings, either as bytes or strings.

        Returns
        -------
        np.ndarray
            A numpy array containing the tokenized sequences of SMILES strings.
        """
        if isinstance(ids_b, np.ndarray):
            ids_b = ids_b.tolist()  # Convert ndarray to list
        converted_ids_b = []
        for smiles in ids_b:
            if isinstance(smiles, bytes) and sys.version_info[0] != 2:
                converted_ids_b.append(
                    TextCNNModel.convert_bytes_to_char(smiles))
            elif isinstance(smiles, str):
                converted_ids_b.append(smiles)
            else:
                raise TypeError("Expected bytes or str, received: {}".format(
                    type(smiles)))
        smiles_seqs = [self.smiles_to_seq(smiles) for smiles in converted_ids_b]
        return np.vstack(smiles_seqs)



import warnings
import torch
import numpy as np
from typing import Callable, List
from deepchem.utils.differentiation_utils.optimize.rootsolver import TerminationCondition
from deepchem.utils import ConvergenceWarning


def anderson_acc(
    fcn: Callable[..., torch.Tensor],
    x0: torch.Tensor,
    params: List,
    feat_ndims: int = 1,
    # anderson_acc parameters
    msize: int = 5,  # how many previous iterations we should save
    beta: float = 1.0,
    lmbda: float = 1e-4,  # small number to ensure invertability of the matrix
    # stopping criteria
    maxiter=None,
    f_tol=None,
    f_rtol=None,
    x_tol=None,
    x_rtol=None,
    custom_terminator=None,
    # misc options
    verbose: bool = False,
) -> torch.Tensor:
    """
    Solve the equilibrium (or fixed-point iteration) problem using Anderson acceleration.

    Examples
    --------
    >>> import torch
    >>> def fcn(x):
    ...     return x
    >>> x0 = torch.tensor([0.0], requires_grad=True)
    >>> x = anderson_acc(fcn, x0, [], 2, 10, maxiter=1000)
    >>> x
    tensor([0.], requires_grad=True)

    Parameters
    ----------
    feat_ndims: int
        The number of dimensions at the end that describe the features (i.e. non-batch dimensions)
    msize: int
        The maximum number of previous iterations we should save for the algorithm
    beta: float
        The damped or overcompensated parameters
    lmbda: float
        Small number to ensure invertability of the matrix
    maxiter: int or None
        Maximum number of iterations, or inf if it is set to None.
    f_tol: float or None
        The absolute tolerance of the norm of the output ``f - x``.
    f_rtol: float or None
        The relative tolerance of the norm of the output ``f - x``.
    x_tol: float or None
        The absolute tolerance of the norm of the input ``x``.
    x_rtol: float or None
        The relative tolerance of the norm of the input ``x``.
    verbose: bool
        Options for verbosity

    References
    ----------
    .. [1] H. F. Walker and P. Ni,
           "Anderson Acceleration for Fixed-Point Iterations".
           Siam J. Numer. Anal.
           https://users.wpi.edu/~walker/Papers/Walker-Ni,SINUM,V49,1715-1735.pdf
    """
    # x0: (..., *nfeats)
    featshape = x0.shape[-feat_ndims:]
    batch_shape = x0.shape[:-feat_ndims]
    feat_size = int(np.prod(x0.shape[-feat_ndims:]))
    dtype = x0.dtype
    device = x0.device

    if maxiter is None:
        maxiter = 100 * (feat_size + 1)

    def _ravel(x: torch.Tensor) -> torch.Tensor:
        # reshape x to have a shape of (batch_size, feats_dim)
        # x: (..., *nfeats)
        xn = x.reshape(*batch_shape, -1)  # (..., feats_tot)
        return xn

    def _unravel(xn: torch.Tensor) -> torch.Tensor:
        # xn: (..., feats_tot)
        x = xn.reshape(*batch_shape, *featshape)  # (..., *nfeats)
        return x

    def _fcn(xn: torch.Tensor) -> torch.Tensor:
        # x: (..., feats_dim)
        x = _unravel(xn)
        fn = _ravel(fcn(x, *params))
        return fn  # (..., feats_dim)

    xn = _ravel(x0)
    fn = _fcn(xn)
    xcollect = torch.zeros((*batch_shape, msize, feat_size),
                           dtype=dtype,
                           device=device)
    fcollect = torch.zeros((*batch_shape, msize, feat_size),
                           dtype=dtype,
                           device=device)
    xcollect[..., 0, :] = xn
    fcollect[..., 0, :] = fn
    xn = fn
    fn = _fcn(xn)
    xcollect[..., 1, :] = xn
    fcollect[..., 1, :] = fn

    hmat = torch.zeros((*batch_shape, msize + 1, msize + 1),
                       dtype=dtype,
                       device=device)
    y = torch.zeros((*batch_shape, msize + 1, 1), dtype=dtype, device=device)
    hmat[..., 0, 1:] = 1.0
    hmat[..., 1:, 0] = 1.0
    y[..., 0, :] = 1.0

    devnorm = (fn - xn).norm()
    stop_cond = custom_terminator if custom_terminator is not None \
        else TerminationCondition(f_tol, f_rtol, devnorm, x_tol, x_rtol)
    if devnorm == 0:
        return x0

    converge = False
    for k in range(2, maxiter):
        nsize = min(k, msize)
        g = fcollect[..., :nsize, :] - xcollect[
            ..., :nsize, :]  # (..., nsize, feat_size)
        # torch.bmm(g, g.transpose(-2, -1))
        hmat[..., 1:nsize + 1, 1:nsize + 1] = torch.einsum("...nf,...mf->...nm", g, g) + \
            lmbda * torch.eye(nsize, dtype=dtype, device=device)
        # alpha: (batch_size, nsize)
        alpha = torch.linalg.solve(hmat[..., :nsize + 1, :nsize + 1],
                                   y[..., :nsize + 1, :])[..., 1:nsize + 1, 0]
        xnew = torch.einsum("...n,...nf->...f", alpha, fcollect[..., :nsize, :]) * beta + \
            torch.einsum("...n,...nf->...f", alpha, xcollect[..., :nsize, :]) * (1 - beta)
        fnew = _fcn(xnew)
        xcollect[..., k % msize, :] = xnew
        fcollect[..., k % msize, :] = fnew

        # check the stopping condition
        to_stop = stop_cond.check(xnew, fnew - xnew, xnew - xn)
        if verbose:
            if k < 10 or k % 10 == 0 or to_stop:
                print("%6d: |dx|=%.3e, |f-x|=%.3e" % (k, (xnew - xn).norm(),
                                                      (fnew - xnew).norm()))
        # update the xn
        xn = xnew
        if to_stop:
            converge = True
            break

    if not converge:
        msg = ("The rootfinder does not converge after %d iterations.") % (
            maxiter)
        warnings.warn(ConvergenceWarning(msg))

    return _unravel(xn)



"""Evaluation metrics."""

import numpy as np
import scipy.stats
from sklearn.metrics import matthews_corrcoef  # noqa
from sklearn.metrics import recall_score  # noqa
from sklearn.metrics import cohen_kappa_score
from sklearn.metrics import r2_score  # noqa
from sklearn.metrics import mean_squared_error
from sklearn.metrics import mean_absolute_error
from sklearn.metrics import precision_score  # noqa
from sklearn.metrics import precision_recall_curve
from sklearn.metrics import auc
from sklearn.metrics import jaccard_score
from sklearn.metrics import f1_score
from sklearn.metrics import roc_auc_score  # noqa
from sklearn.metrics import accuracy_score  # noqa
from sklearn.metrics import balanced_accuracy_score  # noqa
from sklearn.metrics import top_k_accuracy_score  # noqa

# kappa_score is an alias for `sklearn.metrics.cohen_kappa_score`
kappa_score = cohen_kappa_score


def pearsonr(y: np.ndarray, y_pred: np.ndarray) -> float:
    """Computes Pearson correlation coefficient.

    Parameters
    ----------
    y: np.ndarray
        ground truth array
    y_pred: np.ndarray
        predicted array

    Returns
    -------
    float
        The Pearson correlation coefficient.
    """
    return scipy.stats.pearsonr(y, y_pred)[0]


def pearson_r2_score(y: np.ndarray, y_pred: np.ndarray) -> float:
    """Computes Pearson R^2 (square of Pearson correlation).

    Parameters
    ----------
    y: np.ndarray
        ground truth array
    y_pred: np.ndarray
        predicted array

    Returns
    -------
    float
        The Pearson-R^2 score.
    """
    return scipy.stats.pearsonr(y, y_pred)[0]**2


def jaccard_index(y: np.ndarray, y_pred: np.ndarray) -> float:
    """Computes Jaccard Index which is the Intersection Over Union metric
    which is commonly used in image segmentation tasks.

    DEPRECATED: WILL BE REMOVED IN A FUTURE VERSION OF DEEEPCHEM. USE `jaccard_score` instead.

    Parameters
    ----------
    y: np.ndarray
        ground truth array
    y_pred: np.ndarray
        predicted array

    Returns
    -------
    score: float
        The jaccard index. A number between 0 and 1.
    """
    return jaccard_score(y, y_pred)


def pixel_error(y: np.ndarray, y_pred: np.ndarray) -> float:
    """An error metric in case y, y_pred are images.

    Defined as 1 - the maximal F-score of pixel similarity, or squared
    Euclidean distance between the original and the result labels.

    Parameters
    ----------
    y: np.ndarray
        ground truth array
    y_pred: np.ndarray
        predicted array

    Returns
    -------
    score: float
        The pixel-error. A number between 0 and 1.
    """
    return 1 - f1_score(y, y_pred)


def prc_auc_score(y: np.ndarray, y_pred: np.ndarray) -> float:
    """Compute area under precision-recall curve

    Parameters
    ----------
    y: np.ndarray
        A numpy array of shape `(N, n_classes)` or `(N,)` with true labels
    y_pred: np.ndarray
        Of shape `(N, n_classes)` with class probabilities.

    Returns
    -------
    float
        The area under the precision-recall curve. A number between 0 and 1.
    """
    precision, recall, _ = precision_recall_curve(y[:, 1], y_pred[:, 1])
    return auc(recall, precision)


def rms_score(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """Computes RMS error."""
    return np.sqrt(mean_squared_error(y_true, y_pred))


def mae_score(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """Computes MAE."""
    return mean_absolute_error(y_true, y_pred)


def bedroc_score(y_true: np.ndarray, y_pred: np.ndarray, alpha: float = 20.0):
    """Compute BEDROC metric.

    BEDROC metric implemented according to Truchon and Bayley that modifies
    the ROC score by allowing for a factor of early recognition.
    Please confirm details from [1]_.

    Parameters
    ----------
    y_true: np.ndarray
        Binary class labels. 1 for positive class, 0 otherwise
    y_pred: np.ndarray
        Predicted labels
    alpha: float, default 20.0
        Early recognition parameter

    Returns
    -------
    float
        Value in [0, 1] that indicates the degree of early recognition

    Notes
    -----
    This function requires RDKit to be installed.

    References
    ----------
    .. [1] Truchon et al. "Evaluating virtual screening methods: good and bad metrics
      for the “early recognition” problem." Journal of chemical information and modeling
      47.2 (2007): 488-508.
    """
    try:
        from rdkit.ML.Scoring.Scoring import CalcBEDROC
    except ModuleNotFoundError:
        raise ImportError("This function requires RDKit to be installed.")

    # validation
    assert len(y_true) == len(y_pred), 'Number of examples do not match'
    assert np.array_equal(np.unique(y_true).astype(int),
                          [0, 1]), ('Class labels must be binary: %s' %
                                    np.unique(y_true))

    yt = np.asarray(y_true)
    yp = np.asarray(y_pred)

    yt = yt.flatten()
    yp = yp[:, 1].flatten()  # Index 1 because one_hot predictions

    scores = list(zip(yt, yp))
    scores = sorted(scores, key=lambda pair: pair[1], reverse=True)

    return CalcBEDROC(scores, 0, alpha)


def concordance_index(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """Compute Concordance index.

    Statistical metric indicates the quality of the predicted ranking.
    Please confirm details from [1]_.

    Parameters
    ----------
    y_true: np.ndarray
        continous value
    y_pred: np.ndarray
        Predicted value

    Returns
    -------
    float
        score between [0,1]

    References
    ----------
    .. [1] Steck, Harald, et al. "On ranking in survival analysis:
      Bounds on the concordance index." Advances in neural information processing systems (2008): 1209-1216.
    """

    idx = np.argsort(y_true)
    y_true = y_true[idx]
    y_pred = y_pred[idx]

    pairs = 0
    correct_pairs = 0.0

    for i in range(len(y_true)):
        true_a = y_true[i]
        pred_a = y_pred[i]

        for j in range(i + 1, len(y_true)):
            true_b = y_true[j]
            pred_b = y_pred[j]
            if true_a != true_b:
                pairs += 1
                if pred_a == pred_b:
                    correct_pairs += 0.5
                elif pred_a < pred_b:
                    correct_pairs += true_a < true_b
                else:
                    correct_pairs += true_a > true_b

    assert pairs > 0, 'No pairs for comparision'

    return correct_pairs / pairs



import torch
import lightning as L  # noqa
from deepchem.models.torch_models import ModularTorchModel, TorchModel


class DCLightningModule(L.LightningModule):
    """DeepChem Lightning Module to be used with Lightning trainer.

    TODO: Add dataloader, example code and fit, once datasetmodule
    is ready
    The lightning module is a wrapper over deepchem's torch model.
    This module directly works with pytorch lightning trainer
    which runs training for multiple epochs and also is responsible
    for setting up and training models on multiple GPUs.
    https://pytorch-lightning.readthedocs.io/en/stable/api/pytorch_lightning.core.LightningModule.html?highlight=LightningModule

    Notes
    -----
    This class requires PyTorch to be installed.
    """

    def __init__(self, dc_model):
        """Create a new DCLightningModule.

        Parameters
        ----------
        dc_model: deepchem.models.torch_models.torch_model.TorchModel
            TorchModel to be wrapped inside the lightning module.
        """
        super().__init__()
        self.dc_model = dc_model

        self.pt_model = self.dc_model.model
        # This will not work for ModularTorchModel as it is directly uses `loss_func` to compute loss.
        self.loss = self.dc_model._loss_fn

    def configure_optimizers(self):
        return self.dc_model.optimizer._create_pytorch_optimizer(
            self.pt_model.parameters(),)

    def training_step(self, batch, batch_idx):
        """Perform a training step.

        Parameters
        ----------
        batch: A tensor, tuple or list.
        batch_idx: Integer displaying index of this batch
        optimizer_idx: When using multiple optimizers, this argument will also be present.

        Returns
        -------
        loss_outputs: outputs of losses.
        """
        batch = batch.batch_list
        inputs, labels, weights = self.dc_model._prepare_batch(batch)
        if isinstance(inputs, list):
            assert len(inputs) == 1
            inputs = inputs[0]

        if isinstance(self.dc_model, ModularTorchModel):
            loss = self.dc_model.loss_func(inputs, labels, weights)
        elif isinstance(self.dc_model, TorchModel):
            outputs = self.pt_model(inputs)
            if isinstance(outputs, torch.Tensor):
                outputs = [outputs]

            if self.dc_model._loss_outputs is not None:
                outputs = [outputs[i] for i in self.dc_model._loss_outputs]
            loss = self.loss(outputs, labels, weights)

        self.log(
            "train_loss",
            loss,
            on_epoch=True,
            sync_dist=True,
            reduce_fx="mean",
            prog_bar=True,
            batch_size=self.dc_model.batch_size,
        )

        return loss



from deepchem.feat import MolecularFeaturizer
from rdkit import Chem
import numpy as np
from deepchem.feat.graph_data import GraphData

allowable_features = {
    'possible_atomic_num_list':
        list(range(0, 119)),  # 0 represents a masked atom
    'possible_formal_charge_list': [-5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5],
    'possible_chirality_list': [
        Chem.rdchem.ChiralType.CHI_UNSPECIFIED,
        Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CW,
        Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CCW,
        Chem.rdchem.ChiralType.CHI_OTHER
    ],  # noqa: E122
    'possible_hybridization_list': [
        Chem.rdchem.HybridizationType.S, Chem.rdchem.HybridizationType.SP,
        Chem.rdchem.HybridizationType.SP2, Chem.rdchem.HybridizationType.SP3,
        Chem.rdchem.HybridizationType.SP3D, Chem.rdchem.HybridizationType.SP3D2,
        Chem.rdchem.HybridizationType.UNSPECIFIED
    ],  # noqa: E122
    'possible_numH_list': [0, 1, 2, 3, 4, 5, 6, 7, 8],
    'possible_implicit_valence_list': [0, 1, 2, 3, 4, 5, 6],
    'possible_degree_list': [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
    # 0 represents a masked bond
    'possible_bonds': [
        0, Chem.rdchem.BondType.SINGLE, Chem.rdchem.BondType.DOUBLE,
        Chem.rdchem.BondType.TRIPLE, Chem.rdchem.BondType.AROMATIC
    ],  # noqa: E122
    'possible_bond_dirs': [  # only for double bond stereo information
        Chem.rdchem.BondDir.NONE, Chem.rdchem.BondDir.ENDUPRIGHT,
        Chem.rdchem.BondDir.ENDDOWNRIGHT
    ]  # noqa: E122
}


class SNAPFeaturizer(MolecularFeaturizer):
    """
    This featurizer is based on the SNAP featurizer used in the paper [1].

    Example
    -------
    >>> smiles = ["CC(=O)C"]
    >>> featurizer = SNAPFeaturizer()
    >>> print(featurizer.featurize(smiles))
    [GraphData(node_features=[4, 2], edge_index=[2, 6], edge_features=[6, 2])]

    References
    ----------

    .. [1] Hu, W. et al. Strategies for Pre-training Graph Neural Networks. Preprint at https://doi.org/10.48550/arXiv.1905.12265 (2020).

    """

    def _featurize(self, mol, **kwargs):
        """
        Converts rdkit mol object to the deepchem Graph Data object. Uses
        simplified atom and bond features, represented as indices.

        Parameters
        ----------
        mol: RDKitMol
            RDKit molecule object

        Returns
        -------
        data: GraphData
            Graph data object with the attributes: x, edge_index, edge_features

        """
        # atoms
        atom_features_list = []
        for atom in mol.GetAtoms():
            atom_feature = [
                allowable_features['possible_atomic_num_list'].index(
                    atom.GetAtomicNum())
            ] + [
                allowable_features['possible_chirality_list'].index(
                    atom.GetChiralTag())
            ]
            atom_features_list.append(atom_feature)
        x = np.array(atom_features_list)

        # bonds
        num_bond_features = 2  # bond type, bond direction
        if len(mol.GetBonds()) > 0:  # mol has bonds
            edges_list = []
            edge_features_list = []
            for bond in mol.GetBonds():
                i = bond.GetBeginAtomIdx()
                j = bond.GetEndAtomIdx()
                edge_feature = [
                    allowable_features['possible_bonds'].index(
                        bond.GetBondType())
                ] + [
                    allowable_features['possible_bond_dirs'].index(
                        bond.GetBondDir())
                ]
                edges_list.append((i, j))
                edge_features_list.append(edge_feature)
                edges_list.append((j, i))
                edge_features_list.append(edge_feature)

            # Graph connectivity in COO format with shape [2, num_edges]
            edge_index = np.array(edges_list).T

            # Edge feature matrix with shape [num_edges, num_edge_features]
            edge_feats = np.array(edge_features_list)
        else:  # mol has no bonds
            edge_index = np.empty((2, 0), dtype=np.int8)
            edge_feats = np.empty((0, num_bond_features), dtype=np.int8)

        data = GraphData(node_features=x,
                         edge_index=edge_index,
                         edge_features=edge_feats)

        return data



import pytest
try:
    import torch
except:
    pass


@pytest.mark.torch
def test_safepow():
    from deepchem.utils.safeops_utils import safepow

    a = torch.tensor([1e-35, 2e-40])
    p = torch.tensor([2., 3])
    result = safepow(a, p)
    assert torch.allclose(torch.tensor([1.0000e-24, 1.0000e-36]), result)
    assert torch.allclose(torch.tensor([0., 0.]), a**p)


@pytest.mark.torch
def test_safenorm():
    from deepchem.utils.safeops_utils import safenorm

    a = torch.tensor([1e-35, 2e-40])
    result = safenorm(a, 0)
    assert torch.allclose(torch.tensor(1.4142e-15), result)
    assert torch.allclose(torch.tensor(1.4142e-15), a.norm())


@pytest.mark.torch
def test_occnumber():
    from deepchem.utils.safeops_utils import occnumber

    result = occnumber(torch.tensor(2.6), 3, torch.double, torch.device('cpu'))
    assert torch.allclose(
        torch.tensor([1.0000, 1.0000, 0.6000], dtype=torch.float64), result)


@pytest.mark.torch
def test_construct_occ_number():
    from deepchem.utils.safeops_utils import _construct_occ_number

    result = _construct_occ_number(2.5, 2, 3, 3, torch.double,
                                   torch.device('cpu'))
    assert torch.allclose(
        torch.tensor([1.0000, 1.0000, 0.5000], dtype=torch.float64), result)


@pytest.mark.torch
def test_get_floor_and_ceil():
    from deepchem.utils.safeops_utils import get_floor_and_ceil

    result = get_floor_and_ceil(2.5)
    assert result[0] == 2
    assert result[1] == 3


@pytest.mark.torch
def test_safe_cdist():
    from deepchem.utils.safeops_utils import safe_cdist

    a = torch.tensor([[1., 2], [3, 4]])
    b = torch.tensor([[1., 2], [3, 4]])

    result_1 = safe_cdist(a, b)
    assert torch.allclose(result_1,
                          torch.tensor([[0.0000, 2.8284], [2.8284, 0.0000]]))

    result_2 = safe_cdist(a, b, add_diag_eps=True)
    assert torch.allclose(
        result_2,
        torch.tensor([[1.4142e-12, 2.8284e+00], [2.8284e+00, 1.4142e-12]]))



# flake8: noqa
from deepchem.feat.vocabulary_builders.grover_vocab import (
    GroverAtomVocabularyBuilder, GroverBondVocabularyBuilder,
    GroverAtomVocabTokenizer, GroverBondVocabTokenizer)
from deepchem.feat.vocabulary_builders.hf_vocab import HuggingFaceVocabularyBuilder



import unittest
try:
    import torch
    has_torch = True
except ModuleNotFoundError:
    has_torch = False

if has_torch:
    from deepchem.utils import equivariance_utils


class TestEquivarianceUtils(unittest.TestCase):

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_su2_generators_shape(self) -> None:
        # Test if the output has the correct shape.
        j_values = [1, 2, 3, 5,
                    7]  # Test for multiple quantum angular momentum values
        for j in j_values:
            with self.subTest(j=j):
                generators = equivariance_utils.su2_generators(j)
                self.assertEqual(generators.shape, (3, 2 * j + 1, 2 * j + 1))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_su2_generators_zero_momenta(self) -> None:
        # Test for the case of zero momentum (j=0).
        j = 0
        generators = equivariance_utils.su2_generators(j)
        expected_generators = torch.zeros((3, 1, 1), dtype=torch.complex64)
        self.assertTrue(torch.allclose(generators, expected_generators))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_su2_generators_angular_momentum_one(self) -> None:
        # Test for the case of momentum j=1 (spin-1).
        j = 1
        generators = equivariance_utils.su2_generators(j)
        # Expected J_x, J_z, J_y matrices for j=1

        expected_generators = torch.tensor(
            [[[0.0000 + 0.0000j, 0.7071 + 0.0000j, 0.0000 + 0.0000j],
              [-0.7071 + 0.0000j, 0.0000 + 0.0000j, 0.7071 + 0.0000j],
              [0.0000 + 0.0000j, -0.7071 + 0.0000j, 0.0000 + 0.0000j]],
             [[-0.0000 - 1.0000j, 0.0000 + 0.0000j, 0.0000 + 0.0000j],
              [0.0000 + 0.0000j, 0.0000 + 0.0000j, 0.0000 + 0.0000j],
              [0.0000 + 0.0000j, 0.0000 + 0.0000j, 0.0000 + 1.0000j]],
             [[0.0000 - 0.0000j, 0.0000 + 0.7071j, 0.0000 - 0.0000j],
              [0.0000 + 0.7071j, 0.0000 - 0.0000j, 0.0000 + 0.7071j],
              [0.0000 - 0.0000j, 0.0000 + 0.7071j, 0.0000 - 0.0000j]]])

        self.assertTrue(torch.allclose(generators, expected_generators))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_su2_commutation(self):
        j_values = [0, 0.5, 1, 1.5, 2,
                    2.5]  # Test for multiple quantum angular momentum values
        for j in j_values:
            with self.subTest(j=j):
                X = equivariance_utils.su2_generators(j)
                self.assertTrue(
                    torch.allclose(equivariance_utils.commutator(X[0], X[1]),
                                   X[2]))
                self.assertTrue(
                    torch.allclose(equivariance_utils.commutator(X[1], X[2]),
                                   X[0]))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_change_basis_real_to_complex_j_0(self):
        # Test for j = 0, which means we have a 1x1 transformation matrix
        j = 0
        Q = equivariance_utils.change_basis_real_to_complex(
            j, dtype=torch.complex128)
        expected_Q = torch.tensor([[1]], dtype=torch.complex128)
        self.assertTrue(torch.allclose(Q, expected_Q))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_change_basis_real_to_complex_j_2(self):
        # Test for j = 2, which means we have a 5x5 transformation matrix
        j = 2
        Q = equivariance_utils.change_basis_real_to_complex(
            j, dtype=torch.complex64)
        expected_Q = torch.tensor(
            [[
                0.0000 + 0.7071j, -0.0000 + 0.0000j, -0.0000 + 0.0000j,
                -0.0000 + 0.0000j, -0.7071 + 0.0000j
            ],
             [
                 -0.0000 + 0.0000j, 0.0000 + 0.7071j, -0.0000 + 0.0000j,
                 -0.7071 + 0.0000j, -0.0000 + 0.0000j
             ],
             [
                 -0.0000 + 0.0000j, -0.0000 + 0.0000j, -1.0000 + 0.0000j,
                 -0.0000 + 0.0000j, -0.0000 + 0.0000j
             ],
             [
                 -0.0000 + 0.0000j, 0.0000 + 0.7071j, -0.0000 + 0.0000j,
                 0.7071 - 0.0000j, -0.0000 + 0.0000j
             ],
             [
                 -0.0000 - 0.7071j, -0.0000 + 0.0000j, -0.0000 + 0.0000j,
                 -0.0000 + 0.0000j, -0.7071 + 0.0000j
             ]],
            dtype=torch.complex64)
        self.assertTrue(torch.allclose(Q, expected_Q))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_change_basis_real_to_complex_device(self) -> None:
        # Test for device placement (CPU to CUDA)
        j = 1
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        Q = equivariance_utils.change_basis_real_to_complex(j, device=device)
        self.assertEqual(Q.device, device)

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_change_basis_real_to_complex_dtype_conversion(self):
        # Test for dtype conversion (complex128 to complex64)
        j = 1
        Q_complex128 = equivariance_utils.change_basis_real_to_complex(
            j, dtype=torch.complex128)
        Q_complex64 = equivariance_utils.change_basis_real_to_complex(
            j, dtype=torch.complex64)

        self.assertEqual(Q_complex128.dtype, torch.complex128)
        self.assertEqual(Q_complex64.dtype, torch.complex64)

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_so3_generators_shape(self):
        j_values = [1, 2, 3, 4, 5]
        for j in j_values:
            with self.subTest(j=j):
                result = equivariance_utils.so3_generators(j)
                expected_shape = (3, 2 * j + 1, 2 * j + 1)
                self.assertEqual(result.shape, expected_shape)

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_so3_commutation(self):
        j_values = [0, 1, 2, 3, 4,
                    5]  # Test for multiple quantum angular momentum values
        for j in j_values:
            with self.subTest(j=j):
                X = equivariance_utils.so3_generators(j)
                self.assertTrue(
                    torch.allclose(equivariance_utils.commutator(X[0], X[1]),
                                   X[2]))
                self.assertTrue(
                    torch.allclose(equivariance_utils.commutator(X[1], X[2]),
                                   X[0]))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_so3_j0(self):
        j = 0
        result = equivariance_utils.so3_generators(j)
        expected = torch.tensor([[[0.]], [[0.]], [[0.]]],
                                dtype=torch.float64).float()
        self.assertTrue(torch.allclose(result, expected))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_so3_j1(self):
        j = 1
        result = equivariance_utils.so3_generators(j)
        expected = torch.tensor(
            [[[0.0000, 0.0000, 0.0000], [0.0000, 0.0000, -1.0000],
              [0.0000, 1.0000, 0.0000]],
             [[0.0000, 0.0000, 1.0000], [0.0000, 0.0000, 0.0000],
              [-1.0000, 0.0000, 0.0000]],
             [[0.0000, -1.0000, 0.0000], [1.0000, 0.0000, 0.0000],
              [0.0000, 0.0000, 0.0000]]],
            dtype=torch.float64).float()
        self.assertTrue(torch.allclose(result, expected, atol=1e-5))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_unitary_property(self):
        j = 2
        alpha = torch.tensor([0.2])
        beta = torch.tensor([0.1])
        gamma = torch.tensor([0.7])

        D_matrix = equivariance_utils.wigner_D(j, alpha, beta, gamma)
        D_matrix = D_matrix[0]
        conjugate_transpose = torch.transpose(torch.conj(D_matrix), 0, 1)
        identity_matrix = torch.eye(D_matrix.shape[0], dtype=D_matrix.dtype)

        self.assertTrue(
            torch.allclose(D_matrix @ conjugate_transpose,
                           identity_matrix,
                           atol=1e-5))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_orthogonality(self):
        j = 2
        alpha = torch.tensor([0.2])
        beta = torch.tensor([0.1])
        gamma = torch.tensor([0.7])

        D_matrix = equivariance_utils.wigner_D(j, alpha, beta, gamma)
        D_matrix = D_matrix[0]
        num_columns = D_matrix.shape[1]

        for col1 in range(num_columns):
            for col2 in range(num_columns):
                if col1 != col2:
                    dot_product = torch.dot(D_matrix[:, col1], D_matrix[:,
                                                                        col2])
                    self.assertAlmostEqual(dot_product.item(), 0.0, places=5)



"""
Derived from: https://github.com/diffqc/dqc/blob/master/dqc/hamilton/base_hamilton.py
"""
from __future__ import annotations
import torch
from abc import abstractmethod, abstractproperty
from typing import List, Optional, Union, Tuple
from deepchem.utils.dft_utils import BaseGrid, BaseXC, BaseDF, SpinParam
from deepchem.utils.differentiation_utils import EditableModule, LinearOperator


class BaseHamilton(EditableModule):
    """
    Hamilton is a class that provides the LinearOperator of the Hamiltonian
    components.

    The Hamiltonian represents the total energy operator for a system of
    interacting electrons. The Kohn-Sham DFT approach introduces a set of
    fictitious non-interacting electrons that move in an effective potential.
    The total energy functional, which includes the kinetic energy of these
    fictitious electrons and their interaction with an effective potential
    (including the electron-electron interaction), is minimized to obtain the
    ground-state electronic structure.

    The Kohn-Sham Hamiltonian is a key component of this approach, representing
    the operator that governs the evolution of the Kohn-Sham orbitals. It
    includes terms for the kinetic energy of electrons, the external potential
    (usually from nuclei), and the exchange-correlation potential that accounts
    for the electron-electron interactions.

    The Fock matrix represents the one-electron part of the Hamiltonian matrix. Its
    components include kinetic energy, nuclear attraction, and electron-electron
    repulsion integrals. The Fock matrix is pivotal in solving the electronic
    Schrödinger equation and determining the electronic structure of molecular
    systems.

    Examples
    --------
    >>> from deepchem.utils.dft_utils import BaseHamilton
    >>> class MyHamilton(BaseHamilton):
    ...    def __init__(self):
    ...        self._nao = 2
    ...        self._kpts = torch.tensor([[0.0, 0.0, 0.0]])
    ...        self._df = None
    ...    @property
    ...    def nao(self):
    ...        return self._nao
    ...    @property
    ...    def kpts(self):
    ...        return self._kpts
    ...    @property
    ...    def df(self):
    ...        return self._df
    ...    def build(self):
    ...        return self
    ...    def get_nuclattr(self):
    ...        return torch.ones((1, 1, self.nao, self.nao))
    >>> ham = MyHamilton()
    >>> hamilton = ham.build()
    >>> hamilton.get_nuclattr()
    tensor([[[[1., 1.],
              [1., 1.]]]])

    """

    # properties
    @abstractproperty
    def nao(self) -> int:
        """Number of atomic orbital basis.

        Returns
        -------
        int
            Number of atomic orbital basis.

        """
        pass

    @abstractproperty
    def kpts(self) -> torch.Tensor:
        """List of k-points in the Hamiltonian.

        Returns
        -------
        torch.Tensor
            List of k-points in the Hamiltonian. Shape: (nkpts, ndim)

        """
        pass

    @abstractproperty
    def df(self) -> Optional[BaseDF]:
        """
        Returns the density fitting object (if any) attached to this
        Hamiltonian object.

        Returns
        -------
        Optional[BaseDF]
            Returns the density fitting object (if any) attached to this
            Hamiltonian object.

        """
        pass

    # setups
    @abstractmethod
    def build(self):
        """
        Construct the elements needed for the Hamiltonian.
        Heavy-lifting operations should be put here.

        """
        pass

    @abstractmethod
    def setup_grid(self, grid: BaseGrid, xc: Optional[BaseXC] = None) -> None:
        """
        Setup the basis (with its grad) in the spatial grid and prepare the
        gradient of atomic orbital according to the ones required by the xc.
        If xc is not given, then only setup the grid with ao (without any
        gradients of ao)

        Parameters
        ----------
        grid: BaseGrid
            Grid used to setup this Hamilton.
        xc: Optional[BaseXC] (default None)
            Exchange Corelation functional of this Hamiltonian.

        """
        pass

    # fock matrix components
    @abstractmethod
    def get_nuclattr(self) -> LinearOperator:
        """LinearOperator of the nuclear Coulomb attraction.

        Nuclear Coulomb attraction is the electrostatic force binding electrons
        to a nucleus. Positively charged protons attract negatively charged
        electrons, creating stability in quantum systems. This force plays a
        fundamental role in determining the structure and behavior of atoms,
        contributing significantly to the overall potential energy in atomic
        physics.

        Returns
        -------
        LinearOperator
            LinearOperator of the nuclear Coulomb attraction. Shape: (`*BH`, nao, nao)

        """
        pass

    @abstractmethod
    def get_kinnucl(self) -> LinearOperator:
        """
        Returns the LinearOperator of the one-electron operator (i.e. kinetic
        and nuclear attraction). Action of a LinearOperator on a function is a
        linear transformation. In the case of one-electron operators, these
        transformations are essential for solving the Schrödinger equation and
        understanding the behavior of electrons in an atomic or molecular system.

        Returns
        -------
        LinearOperator
            LinearOperator of the one-electron operator. Shape: (`*BH`, nao, nao)

        """
        pass

    @abstractmethod
    def get_overlap(self) -> LinearOperator:
        """
        Returns the LinearOperator representing the overlap of the basis.
        The overlap of the basis refers to the degree to which atomic or
        molecular orbitals in a quantum mechanical system share common space.

        Returns
        -------
        LinearOperator
            LinearOperator representing the overlap of the basis.
            Shape: (`*BH`, nao, nao)

        """
        pass

    @abstractmethod
    def get_elrep(self, dm: torch.Tensor) -> LinearOperator:
        """
        Obtains the LinearOperator of the Coulomb electron repulsion operator.
        Known as the J-matrix.

        In the context of electronic structure theory, it accounts for the
        repulsive interaction between electrons in a many-electron system. The
        J-matrix elements involve the Coulombic interactions between pairs of
        electrons, influencing the total energy and behavior of the system.

        Parameters
        ----------
        dm: torch.Tensor
            Density matrix. Shape: (`*BD`, nao, nao)

        Returns
        -------
        LinearOperator
            LinearOperator of the Coulomb electron repulsion operator.
            Shape: (`*BDH`, nao, nao)

        """
        pass

    @abstractmethod
    def get_exchange(
        self, dm: Union[torch.Tensor, SpinParam[torch.Tensor]]
    ) -> Union[LinearOperator, SpinParam[LinearOperator]]:
        """
        Obtains the LinearOperator of the exchange operator.
        It is -0.5 * K where K is the K matrix obtained from 2-electron integral.

        Exchange operator is a mathematical representation of the exchange
        interaction between identical particles, such as electrons. The
        exchange operator quantifies the effect of interchanging the
        positions of two particles.

        Parameters
        ----------
        dm: Union[torch.Tensor, SpinParam[torch.Tensor]]
            Density matrix. Shape: (`*BD`, nao, nao)

        Returns
        -------
        Union[LinearOperator, SpinParam[LinearOperator]]
            LinearOperator of the exchange operator. Shape: (`*BDH`, nao, nao)

        """
        pass

    @abstractmethod
    def get_vext(self, vext: torch.Tensor) -> LinearOperator:
        r"""
        Returns a LinearOperator of the external potential in the grid.

        .. math::
            \mathbf{V}_{ij} = \int b_i(\mathbf{r}) V(\mathbf{r}) b_j(\mathbf{r})\ d\mathbf{r}

        External potential energy that a particle experiences in a discretized
        space or grid. In quantum mechanics or computational physics, when
        solving for the behavior of particles, an external potential is often
        introduced to represent the influence of external forces.

        Parameters
        ----------
        vext: torch.Tensor
            External potential in the grid. Shape: (`*BR`, ngrid)

        Returns
        -------
        LinearOperator
            LinearOperator of the external potential in the grid. Shape: (`*BRH`, nao, nao)

        """
        pass

    @abstractmethod
    def get_vxc(
        self, dm: Union[torch.Tensor, SpinParam[torch.Tensor]]
    ) -> Union[LinearOperator, SpinParam[LinearOperator]]:
        """
        Returns a LinearOperator for the exchange-correlation potential

        The exchange-correlation potential combines two effects:

        1. Exchange potential: Arises from the antisymmetry of the electron
        wave function. It quantifies the tendency of electrons to avoid each
        other due to their indistinguishability.

        2. Correlation potential: Accounts for the electron-electron
        correlation effects that arise from the repulsion between electrons.

        TODO: check if what we need for Meta-GGA involving kinetics and for
        exact-exchange

        Parameters
        ----------
        dm: Union[torch.Tensor, SpinParam[torch.Tensor]]
            Density matrix. Shape: (`*BD`, nao, nao)

        Returns
        -------
        Union[LinearOperator, SpinParam[LinearOperator]]
            LinearOperator for the exchange-correlation potential. Shape: (`*BDH`, nao, nao)

        """
        pass

    # interface to dm
    @abstractmethod
    def ao_orb2dm(self, orb: torch.Tensor,
                  orb_weight: torch.Tensor) -> torch.Tensor:
        """Convert the atomic orbital to the density matrix.

        Parameters
        ----------
        orb: torch.Tensor
            Atomic orbital. Shape: (*BO, nao, norb)
        orb_weight: torch.Tensor
            Orbital weight. Shape: (*BW, norb)

        Returns
        -------
        torch.Tensor
            Density matrix. Shape: (*BOWH, nao, nao)

        """
        pass

    @abstractmethod
    def aodm2dens(self, dm: torch.Tensor, xyz: torch.Tensor) -> torch.Tensor:
        """Get the density value in the Cartesian coordinate.

        Parameters
        ----------
        dm: torch.Tensor
            Density matrix. Shape: (`*BD`, nao, nao)
        xyz: torch.Tensor
            Cartesian coordinate. Shape: (`*BR`, ndim)

        Returns
        -------
        torch.Tensor
            Density value in the Cartesian coordinate. Shape: (`*BRD`)

        """
        pass

    # energy of the Hamiltonian
    @abstractmethod
    def get_e_hcore(self, dm: torch.Tensor) -> torch.Tensor:
        """
        Get the energy from the one-electron Hamiltonian. The input is total
        density matrix.

        Parameters
        ----------
        dm: torch.Tensor
            Total Density matrix.

        Returns
        -------
        torch.Tensor
            Energy from the one-electron Hamiltonian.

        """
        pass

    @abstractmethod
    def get_e_elrep(self, dm: torch.Tensor) -> torch.Tensor:
        """
        Get the energy from the electron repulsion. The input is total density
        matrix.

        Parameters
        ----------
        dm: torch.Tensor
            Total Density matrix.

        Returns
        -------
        torch.Tensor
            Energy from the one-electron Hamiltonian.

        """
        pass

    @abstractmethod
    def get_e_exchange(
            self, dm: Union[torch.Tensor,
                            SpinParam[torch.Tensor]]) -> torch.Tensor:
        """Get the energy from the exact exchange.

        Parameters
        ----------
        dm: Union[torch.Tensor, SpinParam[torch.Tensor]]
            Density matrix.

        Returns
        -------
        torch.Tensor
            Energy from the exact exchange.

        """
        pass

    @abstractmethod
    def get_e_xc(
            self, dm: Union[torch.Tensor,
                            SpinParam[torch.Tensor]]) -> torch.Tensor:
        """
        Returns the exchange-correlation energy using the xc object given in
        ``.setup_grid()``

        Parameters
        ----------
        dm: Union[torch.Tensor, SpinParam[torch.Tensor]]
            Density matrix. Shape: (`*BD`, nao, nao)

        Returns
        -------
        torch.Tensor
            Exchange-correlation energy.

        """
        pass

    # free parameters for variational method
    @abstractmethod
    def ao_orb_params2dm(
            self,
            ao_orb_params: torch.Tensor,
            ao_orb_coeffs: torch.Tensor,
            orb_weight: torch.Tensor,
            with_penalty: Optional[float] = None) -> List[torch.Tensor]:
        """
        Convert the atomic orbital free parameters (parametrized in such a way
        so it is not bounded) to the density matrix.

        Parameters
        ----------
        ao_orb_params: torch.Tensor
            The tensor that parametrized atomic orbital in an unbounded space.
        ao_orb_coeffs: torch.Tensor
            The tensor that helps ``ao_orb_params`` in describing the orbital.
            The difference with ``ao_orb_params`` is that ``ao_orb_coeffs`` is
            not differentiable and not to be optimized in variational method.
        orb_weight: torch.Tensor
            The orbital weights.
        with_penalty: float or None
            If a float, it returns a tuple of tensors where the first element is
            ``dm``, and the second element is the penalty multiplied by the
            penalty weights. The penalty is to compensate the overparameterization
            of ``ao_orb_params``, stabilizing the Hessian for gradient calculation.

        Returns
        -------
        torch.Tensor or tuple of torch.Tensor
            The density matrix from the orbital parameters and (if ``with_penalty``)
            the penalty of the overparameterization of ``ao_orb_params``.

        Notes
        -----
        * The penalty should be 0 if ``ao_orb_params`` is from ``dm2ao_orb_params``.
        * The density matrix should be recoverable when put through ``dm2ao_orb_params``
          and ``ao_orb_params2dm``.

        """
        pass

    @abstractmethod
    def dm2ao_orb_params(self, dm: torch.Tensor,
                         norb: int) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Convert from the density matrix to the orbital parameters.
        The map is not one-to-one, but instead one-to-many where there might
        be more than one orbital parameters to describe the same density matrix.
        For restricted systems, only one of the ``dm`` (``dm.u`` or ``dm.d``) is
        sufficient.

        Parameters
        ----------
        dm: torch.Tensor
            The density matrix.
        norb: int
            The number of orbitals for the system.

        Returns
        -------
        tuple of 2 torch.Tensor
            The atomic orbital parameters for the first returned value and the
            atomic orbital coefficients for the second value.

        """
        pass

    # Editable module
    @abstractmethod
    def getparamnames(self, methodname: str, prefix: str = "") -> List[str]:
        """Return the paramnames

        Parameters
        ----------
        methodname: str
            The name of the method.
        prefix: str (default "")
            The prefix of the paramnames.

        Returns
        -------
        List[str]
            The paramnames.

        """
        pass



"""
Density Functional Theory Data Structure Utilities
"""
from typing import Union, TypeVar, Generic, Optional, Callable, List, Dict
from dataclasses import dataclass
import torch
import numpy as np
from deepchem.utils import gaussian_integral

__all__ = ["ZType"]

T = TypeVar('T')
P = TypeVar('P')

# type of the atom Z
ZType = Union[int, float, torch.Tensor]

# input types
AtomZsType = Union[List[str], List[ZType], torch.Tensor]
AtomPosType = Union[List[List[float]], np.ndarray, torch.Tensor]


@dataclass
class SpinParam(Generic[T]):
    """Data structure to store different values for spin-up and spin-down electrons.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.dft_utils import SpinParam
    >>> dens_u = torch.ones(1)
    >>> dens_d = torch.zeros(1)
    >>> sp = SpinParam(u=dens_u, d=dens_d)
    >>> sp.u
    tensor([1.])
    >>> sp.sum()
    tensor([1.])
    >>> sp.reduce(torch.multiply)
    tensor([0.])

    """

    def __init__(self, u: T, d: T):
        """Initialize the SpinParam object.

        Parameters
        ----------
        u: any type
            The parameters that corresponds to the spin-up electrons.
        d: any type
            The parameters that corresponds to the spin-down electrons.

        """
        self.u = u
        self.d = d

    def __repr__(self) -> str:
        """Return the string representation of the SpinParam object."""
        return f"SpinParam(u={self.u}, d={self.d})"

    def sum(self):
        """Returns the sum of up and down parameters."""

        return self.u + self.d

    def reduce(self, fcn: Callable) -> T:
        """Reduce up and down parameters with the given function."""

        return fcn(self.u, self.d)


@dataclass
class ValGrad:
    """Data structure that contains local information about density profiles.
    Data structure used as a umbrella class for density profiles and the
    derivative of the potential w.r.t. density profiles.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.dft_utils import ValGrad
    >>> dens = torch.ones(1)
    >>> grad = torch.zeros(1)
    >>> lapl = torch.ones(1)
    >>> kin = torch.ones(1)
    >>> vg = ValGrad(value=dens, grad=grad, lapl=lapl, kin=kin)
    >>> vg + vg
    ValGrad(value=tensor([2.]), grad=tensor([0.]), lapl=tensor([2.]), kin=tensor([2.]))
    >>> vg * 5
    ValGrad(value=tensor([5.]), grad=tensor([0.]), lapl=tensor([5.]), kin=tensor([5.]))

    """

    def __init__(self,
                 value: torch.Tensor,
                 grad: Optional[torch.Tensor] = None,
                 lapl: Optional[torch.Tensor] = None,
                 kin: Optional[torch.Tensor] = None):
        """Initialize the ValGrad object.

        Parameters
        ----------
        value: torch.Tensor
            Tensors containing the value of the local information.
        grad: torch.Tensor or None
            If tensor, it represents the gradient of the local information with
            shape ``(..., 3)`` where ``...`` should be the same shape as ``value``.
        lapl: torch.Tensor or None
            If tensor, represents the laplacian value of the local information.
            It should have the same shape as ``value``.
        kin: torch.Tensor or None
            If tensor, represents the local kinetic energy density.
            It should have the same shape as ``value``.

        """
        self.value = value
        self.grad = grad
        self.lapl = lapl
        self.kin = kin

    def __add__(self, b):
        """Add two ValGrad objects together."""
        return ValGrad(
            value=self.value + b.value,
            grad=self.grad + b.grad if self.grad is not None else None,
            lapl=self.lapl + b.lapl if self.lapl is not None else None,
            kin=self.kin + b.kin if self.kin is not None else None,
        )

    def __mul__(self, f: Union[float, int, torch.Tensor]):
        """Multiply the ValGrad object with a scalar."""
        if isinstance(f, torch.Tensor):
            assert f.numel(
            ) == 1, "ValGrad multiplication with tensor can only be done with 1-element tensor"

        return ValGrad(
            value=self.value * f,
            grad=self.grad * f if self.grad is not None else None,
            lapl=self.lapl * f if self.lapl is not None else None,
            kin=self.kin * f if self.kin is not None else None,
        )

    def __repr__(self):
        return f"ValGrad(value={self.value}, grad={self.grad}, lapl={self.lapl}, kin={self.kin})"


@dataclass
class CGTOBasis:
    """Data structure that contains information about a contracted gaussian
    type orbital (CGTO).

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.dft_utils import CGTOBasis
    >>> alphas = torch.ones(1)
    >>> coeffs = torch.ones(1)
    >>> cgto = CGTOBasis(angmom=0, alphas=alphas, coeffs=coeffs)
    >>> cgto.wfnormalize_()
    CGTOBasis(angmom=0, alphas=tensor([1.]), coeffs=tensor([2.5265]), normalized=True)

    """

    def __init__(self, angmom: int, alphas: torch.Tensor, coeffs: torch.Tensor):
        """Initialize the CGTOBasis object.

        Parameters
        ----------
        angmom: int
            The angular momentum of the basis.
        alphas: torch.Tensor
            The gaussian exponents of the basis. Shape: (nbasis,)
        coeffs: torch.Tensor
            The coefficients of the basis. Shape: (nbasis,)

        """
        self.angmom = angmom
        self.alphas = alphas
        self.coeffs = coeffs
        self.normalized = False

    def __repr__(self):
        """Return the string representation of the CGTOBasis object.

        Returns
        -------
        angmom: int
            The angular momentum of the basis.
        alphas: torch.Tensor
            The gaussian exponents of the basis. Shape: (nbasis,)
        coeffs: torch.Tensor
            The coefficients of the basis. Shape: (nbasis,)

        """
        return f"CGTOBasis(angmom={self.angmom}, alphas={self.alphas}, coeffs={self.coeffs}, normalized={self.normalized})"

    def wfnormalize_(self) -> "CGTOBasis":
        """Wavefunction normalization

        The normalization is obtained from CINTgto_norm from
        libcint/src/misc.c, or
        https://github.com/sunqm/libcint/blob/b8594f1d27c3dad9034984a2a5befb9d607d4932/src/misc.c#L80

        Please note that the square of normalized wavefunctions do not integrate
        to 1, but e.g. for s: 4*pi, p: (4*pi/3)

        """

        # if the basis has been normalized before, then do nothing
        if self.normalized:
            return self

        coeffs = self.coeffs

        # normalize to have individual gaussian integral to be 1 (if coeff is 1)
        value = gaussian_integral(2 * self.angmom + 2, 2 * self.alphas)
        assert isinstance(value, torch.Tensor)
        coeffs = coeffs / torch.sqrt(value)

        # normalize the coefficients in the basis (because some basis such as
        # def2-svp-jkfit is not normalized to have 1 in overlap)
        ee = self.alphas.unsqueeze(-1) + self.alphas.unsqueeze(
            -2)  # (ngauss, ngauss)
        ee = gaussian_integral(2 * self.angmom + 2, ee)  # type: ignore
        s1 = 1 / torch.sqrt(torch.einsum("a,ab,b", coeffs, ee, coeffs))
        coeffs = coeffs * s1

        self.coeffs = coeffs
        self.normalized = True
        return self


@dataclass
class AtomCGTOBasis:
    """Data structure that contains information about a atom and its contracted
    gaussian type orbital (CGTO).

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.dft_utils import AtomCGTOBasis, CGTOBasis
    >>> alphas = torch.ones(1)
    >>> coeffs = torch.ones(1)
    >>> cgto = CGTOBasis(angmom=0, alphas=alphas, coeffs=coeffs)
    >>> atomcgto = AtomCGTOBasis(atomz=1, bases=[cgto], pos=[[0.0, 0.0, 0.0]])
    >>> atomcgto
    AtomCGTOBasis(atomz=1, bases=[CGTOBasis(angmom=0, alphas=tensor([1.]), coeffs=tensor([1.]), normalized=False)], pos=tensor([[0., 0., 0.]]))

    """

    def __init__(self, atomz: ZType, bases: List[CGTOBasis], pos: AtomPosType):
        """Initialize the AtomCGTOBasis object.

        Parameters
        ----------
        atomz: ZType
            Atomic number of the atom.
        bases: List[CGTOBasis]
            List of CGTOBasis objects.
        pos: AtomPosType
            Position of the atom. Shape: (ndim,)

        """
        self.atomz = atomz
        self.bases = bases
        self.pos = torch.tensor(pos)

    def __repr__(self):
        """Return the string representation of the AtomCGTOBasis object.

        Returns
        -------
        atomz: ZType
            Atomic number of the atom.
        bases: List[CGTOBasis]
            List of CGTOBasis objects.
        pos: AtomPosType
            Position of the atom.

        """
        return f"AtomCGTOBasis(atomz={self.atomz}, bases={self.bases}, pos={self.pos})"


# input basis type
BasisInpType = Union[str, List[CGTOBasis], List[str], List[List[CGTOBasis]],
                     Dict[Union[str, int], Union[List[CGTOBasis], str]]]



from typing import Callable, List, Optional
import numpy as np

from deepchem.utils.typing import RDKitMol
from deepchem.feat.base_classes import MolecularFeaturizer


class MordredDescriptors(MolecularFeaturizer):
    """Mordred descriptors.

    This class computes a list of chemical descriptors using Mordred.
    Please see the details about all descriptors from [1]_, [2]_.

    Attributes
    ----------
    descriptors: List[str]
        List of Mordred descriptor names used in this class.

    References
    ----------
    .. [1] Moriwaki, Hirotomo, et al. "Mordred: a molecular descriptor calculator."
        Journal of cheminformatics 10.1 (2018): 4.
    .. [2] http://mordred-descriptor.github.io/documentation/master/descriptors.html

    Note
    ----
    This class requires Mordred to be installed.

    Examples
    --------
    >>> import deepchem as dc
    >>> smiles = ['CC(=O)OC1=CC=CC=C1C(=O)O']
    >>> featurizer = dc.feat.MordredDescriptors(ignore_3D=True)
    >>> features = featurizer.featurize(smiles)
    >>> type(features[0])
    <class 'numpy.ndarray'>
    >>> features[0].shape
    (1613,)

    """

    def __init__(self, ignore_3D: bool = True):
        """
        Parameters
        ----------
        ignore_3D: bool, optional (default True)
            Whether to use 3D information or not.

        """
        self.ignore_3D = ignore_3D
        self.calc: Optional[Callable] = None
        self.descriptors: Optional[List] = None

    def _featurize(self, datapoint: RDKitMol, **kwargs) -> np.ndarray:
        """
        Calculate Mordred descriptors.

        Parameters
        ----------
        datapoint: rdkit.Chem.rdchem.Mol
            RDKit Mol object

        Returns
        -------
        np.ndarray
            1D array of Mordred descriptors for `mol`.
            If ignore_3D is True, the length is 1613.
            If ignore_3D is False, the length is 1826.

        """
        if 'mol' in kwargs:
            datapoint = kwargs.get("mol")
            raise DeprecationWarning(
                'Mol is being phased out as a parameter, please pass "datapoint" instead.'
            )
        if self.calc is None:
            try:
                from mordred import Calculator, descriptors, is_missing
                self.is_missing = is_missing
                self.calc = Calculator(descriptors, ignore_3D=self.ignore_3D)
                self.descriptors = list(descriptors.__all__)
            except ModuleNotFoundError:
                raise ImportError(
                    "This class requires Mordred to be installed.")

        feature = self.calc(datapoint)
        # convert errors to zero
        feature = [
            0.0 if self.is_missing(val) or isinstance(val, str) else val
            for val in feature
        ]
        return np.asarray(feature)



import torch
from typing import Optional, Sequence, Tuple, Union, Mapping, Any, Callable
from deepchem.utils.differentiation_utils import LinearOperator, get_bcasted_dims, MatrixLinearOperator, set_default_option, get_and_pop_keys, dummy_context_manager, get_method, solve  # type: ignore
import functools
from deepchem.utils.pytorch_utils import tallqr, to_fortran_order
import warnings


def lsymeig(A: LinearOperator,
            neig: Optional[int] = None,
            M: Optional[LinearOperator] = None,
            bck_options: Mapping[str, Any] = {},
            method: Union[str, Callable, None] = None,
            **fwd_options) -> Tuple[torch.Tensor, torch.Tensor]:
    """Obtain ``neig`` lowest eigenvalues and eigenvectors of a linear operator"""
    return symeig(A,
                  neig,
                  "lowest",
                  M,
                  method=method,
                  bck_options=bck_options,
                  **fwd_options)


def usymeig(A: LinearOperator,
            neig: Optional[int] = None,
            M: Optional[LinearOperator] = None,
            bck_options: Mapping[str, Any] = {},
            method: Union[str, Callable, None] = None,
            **fwd_options) -> Tuple[torch.Tensor, torch.Tensor]:
    """Obtain ``neig`` uppest eigenvalues and eigenvectors of a linear operator"""
    return symeig(A,
                  neig,
                  "uppest",
                  M,
                  method=method,
                  bck_options=bck_options,
                  **fwd_options)


def symeig(A: LinearOperator,
           neig: Optional[int] = None,
           mode: str = "lowest",
           M: Optional[LinearOperator] = None,
           bck_options: Mapping[str, Any] = {},
           method: Union[str, Callable, None] = None,
           **fwd_options) -> Tuple[torch.Tensor, torch.Tensor]:
    r"""
    Obtain ``neig`` lowest eigenvalues and eigenvectors of a linear operator,

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.differentiation_utils import LinearOperator
    >>> A = LinearOperator.m(torch.tensor([[3, -1j], [1j, 4]]))
    >>> evals, evecs = symeig(A)
    >>> evals.shape
    torch.Size([2])
    >>> evecs.shape
    torch.Size([2, 2])

    .. math::

        \mathbf{AX = MXE}

    where :math:`\mathbf{A}, \mathbf{M}` are linear operators,
    :math:`\mathbf{E}` is a diagonal matrix containing the eigenvalues, and
    :math:`\mathbf{X}` is a matrix containing the eigenvectors.
    This function can handle derivatives for degenerate cases by setting non-zero
    ``degen_atol`` and ``degen_rtol`` in the backward option using the expressions
    in [1]_.

    Parameters
    ----------
    A: LinearOperator
        The linear operator object on which the eigenpairs are constructed.
        It must be a Hermitian linear operator with shape ``(*BA, q, q)``
    neig: int or None
        The number of eigenpairs to be retrieved. If ``None``, all eigenpairs are
        retrieved
    mode: str
        ``"lowest"`` or ``"uppermost"``/``"uppest"``. If ``"lowest"``,
        it will take the lowest ``neig`` eigenpairs.
        If ``"uppest"``, it will take the uppermost ``neig``.
    M: LinearOperator
        The transformation on the right hand side. If ``None``, then ``M=I``.
        If specified, it must be a Hermitian with shape ``(*BM, q, q)``.
    bck_options: dict
        Method-specific options for :func:`solve` which used in backpropagation
        calculation with some additional arguments for computing the backward
        derivatives:

        * ``degen_atol`` (``float`` or None): Minimum absolute difference between
          two eigenvalues to be treated as degenerate. If None, it is
          ``torch.finfo(dtype).eps**0.6``. If 0.0, no special treatment on
          degeneracy is applied. (default: None)
        * ``degen_rtol`` (``float`` or None): Minimum relative difference between
          two eigenvalues to be treated as degenerate. If None, it is
          ``torch.finfo(dtype).eps**0.4``. If 0.0, no special treatment on
          degeneracy is applied. (default: None)

        Note: the default values of ``degen_atol`` and ``degen_rtol`` are going
        to change in the future. So, for future compatibility, please specify
        the specific values.

    method: str or callable or None
        Method for the eigendecomposition. If ``None``, it will choose
        ``"exacteig"``.
    **fwd_options
        Method-specific options (see method section below).

    Returns
    -------
    tuple of tensors (eigenvalues, eigenvectors)
        It will return eigenvalues and eigenvectors with shapes respectively
        ``(*BAM, neig)`` and ``(*BAM, na, neig)``, where ``*BAM`` is the
        broadcasted shape of ``*BA`` and ``*BM``.

    References
    ----------
    .. [1] Muhammad F. Kasim,
           "Derivatives of partial eigendecomposition of a real symmetric matrix for degenerate cases".
           arXiv:2011.04366 (2020)
           `https://arxiv.org/abs/2011.04366 <https://arxiv.org/abs/2011.04366>`_

    """
    assert A.is_hermitian, "The linear operator A must be Hermitian"
    assert not torch.is_grad_enabled() or A.is_getparamnames_implemented, \
        "The _getparamnames(self, prefix) of linear operator A must be "\
        "implemented if using symeig with grad enabled"
    if M is not None:
        assert M.is_hermitian, "The linear operator M must be Hermitian"
        assert M.shape[-1] == A.shape[
            -1], "The shape of A & M must match (A: %s, M: %s)" % (A.shape,
                                                                   M.shape)
        assert not torch.is_grad_enabled() or M.is_getparamnames_implemented, \
            "The _getparamnames(self, prefix) of linear operator M must be "\
            "implemented if using symeig with grad enabled"
    mode = mode.lower()
    if mode == "uppermost":
        mode = "uppest"
    if method is None:
        if isinstance(A, MatrixLinearOperator) and \
           (M is None or isinstance(M, MatrixLinearOperator)):
            method = "exacteig"
        else:
            # TODO: implement robust LOBPCG and put it here
            method = "exacteig"
    if neig is None:
        neig = A.shape[-1]

    if method == "exacteig":
        return exacteig(A, neig, mode, M)
    else:
        fwd_options["method"] = method
        # get the unique parameters of A & M
        params = A.getlinopparams()
        mparams = M.getlinopparams() if M is not None else []
        na = len(params)
        return symeig_torchfcn.apply(A, neig, mode, M, fwd_options, bck_options,
                                     na, *params, *mparams)


def svd(A: LinearOperator,
        k: Optional[int] = None,
        mode: str = "uppest",
        bck_options: Mapping[str, Any] = {},
        method: Union[str, Callable, None] = None,
        **fwd_options) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    r"""
    Perform the singular value decomposition (SVD):

    Examples
    --------
    >>> from deepchem.utils.differentiation_utils import svd
    >>> import torch
    >>> from deepchem.utils.differentiation_utils import LinearOperator
    >>> A = LinearOperator.m(torch.tensor([[3, 1], [1, 4.]]))
    >>> svd(A, mode="lowest")
    (tensor([[-0.8507,  0.5257],
            [ 0.5257,  0.8507]]), tensor([2.3820, 4.6180]), tensor([[-0.8507,  0.5257],
            [ 0.5257,  0.8507]]))

    .. math::

        \mathbf{A} = \mathbf{U\Sigma V}^H

    where :math:`\mathbf{U}` and :math:`\mathbf{V}` are semi-unitary matrix and
    :math:`\mathbf{\Sigma}` is a diagonal matrix containing real non-negative
    numbers.
    This function can handle derivatives for degenerate singular values by setting non-zero
    ``degen_atol`` and ``degen_rtol`` in the backward option using the expressions
    in [1]_.

    Parameters
    ----------
    A: LinearOperator
        The linear operator to be decomposed. It has a shape of ``(*BA, m, n)``
        where ``(*BA)`` is the batched dimension of ``A``.
    k: int or None
        The number of decomposition obtained. If ``None``, it will be
        ``min(*A.shape[-2:])``
    mode: str
        ``"lowest"`` or ``"uppermost"``/``"uppest"``. If ``"lowest"``,
        it will take the lowest ``k`` decomposition.
        If ``"uppest"``, it will take the uppermost ``k``.
    bck_options: dict
        Method-specific options for :func:`solve` which used in backpropagation
        calculation with some additional arguments for computing the backward
        derivatives:

        * ``degen_atol`` (``float`` or None): Minimum absolute difference between
          two singular values to be treated as degenerate. If None, it is
          ``torch.finfo(dtype).eps**0.6``. If 0.0, no special treatment on
          degeneracy is applied. (default: None)
        * ``degen_rtol`` (``float`` or None): Minimum relative difference between
          two singular values to be treated as degenerate. If None, it is
          ``torch.finfo(dtype).eps**0.4``. If 0.0, no special treatment on
          degeneracy is applied. (default: None)

        Note: the default values of ``degen_atol`` and ``degen_rtol`` are going
        to change in the future. So, for future compatibility, please specify
        the specific values.

    method: str or callable or None
        Method for the svd (same options for :func:`symeig`). If ``None``,
        it will choose ``"exacteig"``.
    **fwd_options
        Method-specific options (see method section below).

    Returns
    -------
    tuple of tensors (u, s, vh)
        It will return ``u, s, vh`` with shapes respectively
        ``(*BA, m, k)``, ``(*BA, k)``, and ``(*BA, k, n)``.

    Note
    ----
    It is a naive implementation of symmetric eigendecomposition of ``A.H @ A``
    or ``A @ A.H`` (depending which one is cheaper)

    References
    ----------
    .. [1] Muhammad F. Kasim,
           "Derivatives of partial eigendecomposition of a real symmetric matrix for degenerate cases".
           arXiv:2011.04366 (2020)
           `https://arxiv.org/abs/2011.04366 <https://arxiv.org/abs/2011.04366>`_

    """
    # adapted from scipy.sparse.linalg.svds

    m = A.shape[-2]
    n = A.shape[-1]
    if m < n:
        AAsym = A.matmul(A.H, is_hermitian=True)
    else:
        AAsym = A.H.matmul(A, is_hermitian=True)

    eivals, eivecs = symeig(AAsym,
                            k,
                            mode,
                            bck_options=bck_options,
                            method=method,
                            **fwd_options)  # (*BA, k) and (*BA, min(mn), k)

    # clamp the eigenvalues to a small positive values to avoid numerical
    # instability
    eivals = torch.clamp(eivals, min=0.0)
    s = torch.sqrt(eivals)  # (*BA, k)
    sdiv = torch.clamp(s, min=1e-12).unsqueeze(-2)  # (*BA, 1, k)
    if m < n:
        u = eivecs  # (*BA, m, k)
        v = A.rmm(u) / sdiv  # (*BA, n, k)
    else:
        v = eivecs  # (*BA, n, k)
        u = A.mm(v) / sdiv  # (*BA, m, k)
    vh = v.transpose(-2, -1).conj()
    return u, s, vh


class symeig_torchfcn(torch.autograd.Function):
    """A wrapper for symeig to be used in torch.autograd.Function"""

    @staticmethod
    def forward(ctx, A, neig, mode, M, fwd_options, bck_options, na, *amparams):
        """Calculate the eigenvalues and eigenvectors of a linear operator

        Parameters
        ----------
        A: LinearOperator
            The linear operator object on which the eigenpairs are constructed.
            It must be a Hermitian linear operator with shape ``(*BA, q, q)``
        neig: int
            The number of eigenpairs to be retrieved. If ``None``, all eigenpairs are
            retrieved
        mode: str
            ``"lowest"`` or ``"uppermost"``/``"uppest"``. If ``"lowest"``,
            it will take the lowest ``neig`` eigenpairs.
            If ``"uppest"``, it will take the uppermost ``neig``.
        M: xitorch.LinearOperator
            The transformation on the right hand side. If ``None``, then ``M=I``.
            If specified, it must be a Hermitian with shape ``(*BM, q, q)``.
        fwd_options: dict
            Method-specific options (see method section below).
        bck_options: dict
            Method-specific options for :func:`solve` which used in backpropagation
            calculation with some additional arguments for computing the backward
            derivatives: ``degen_atol`` and ``degen_rtol``.
        na: int
            Number of parameters of A (and M if M is not None)
        *amparams: torch.Tensor
            Parameters of A (and M if M is not None)

        """

        # separate the sets of parameters
        params = amparams[:na]
        mparams = amparams[na:]

        config = set_default_option({}, fwd_options)
        ctx.bck_config = set_default_option(
            {
                "degen_atol": None,
                "degen_rtol": None,
            }, bck_options)

        # options for calculating the backward (not for `solve`)
        alg_keys = ["degen_atol", "degen_rtol"]
        ctx.bck_alg_config = get_and_pop_keys(ctx.bck_config, alg_keys)

        method = config.pop("method")
        with A.uselinopparams(*params), M.uselinopparams(
                *mparams) if M is not None else dummy_context_manager():
            methods = {
                "davidson": davidson,
                "exacteig": exacteig,
            }
            method_fcn = get_method("symeig", methods, method)
            evals, evecs = method_fcn(A, neig, mode, M, **config)

        # save for the backward
        # evals: (*BAM, neig)
        # evecs: (*BAM, na, neig)
        ctx.save_for_backward(evals, evecs, *amparams)
        ctx.na = na
        ctx.A = A
        ctx.M = M
        return evals, evecs

    @staticmethod
    def backward(ctx, grad_evals, grad_evecs):
        """Calculate the gradient of the eigenvalues and eigenvectors of a linear operator

        Parameters
        ----------
        grad_evals: torch.Tensor
            The gradient of the eigenvalues. Shape: ``(*BAM, neig)``
        grad_evecs: torch.Tensor
            The gradient of the eigenvectors. Shape: ``(*BAM, na, neig)``

        """

        # get the variables from ctx
        evals, evecs = ctx.saved_tensors[:2]
        na = ctx.na
        amparams = ctx.saved_tensors[2:]
        params = amparams[:na]
        mparams = amparams[na:]

        M = ctx.M
        A = ctx.A
        degen_atol: Optional[float] = ctx.bck_alg_config[
            "degen_atol"]  # type: ignore
        degen_rtol: Optional[float] = ctx.bck_alg_config[
            "degen_rtol"]  # type: ignore

        # set the default values of degen_*tol
        dtype = evals.dtype
        if degen_atol is None:
            degen_atol = torch.finfo(dtype).eps**0.6
        if degen_rtol is None:
            degen_rtol = torch.finfo(dtype).eps**0.4

        # check the degeneracy
        if degen_atol > 0 or degen_rtol > 0:
            # idx_degen: (*BAM, neig, neig)
            idx_degen, isdegenerate = _check_degen(evals, degen_atol,
                                                   degen_rtol)
        else:
            isdegenerate = False
        if not isdegenerate:
            idx_degen = None

        # the loss function where the gradient will be retrieved
        # warnings: if not all params have the connection to the output of A,
        # it could cause an infinite loop because pytorch will keep looking
        # for the *params node and propagate further backward via the `evecs`
        # path. So make sure all the *params are all connected in the graph.
        with torch.enable_grad():
            params = [p.clone().requires_grad_() for p in params]
            with A.uselinopparams(*params):
                loss = A.mm(evecs)  # (*BAM, na, neig)

        # if degenerate, check the conditions for finite derivative
        if isdegenerate:
            xtg = torch.matmul(evecs.transpose(-2, -1).conj(), grad_evecs)
            req1 = idx_degen * (xtg - xtg.transpose(-2, -1).conj())
            reqtol = xtg.abs().max() * grad_evecs.shape[-2] * torch.finfo(
                grad_evecs.dtype).eps

            if not torch.all(torch.abs(req1) <= reqtol):
                # if the requirements are not satisfied, raises a warning
                msg = (
                    "Degeneracy appears but the loss function seem to depend "
                    "strongly on the eigenvector. The gradient might be incorrect.\n"
                )
                msg += "Eigenvalues:\n%s\n" % str(evals)
                msg += "Degenerate map:\n%s\n" % str(idx_degen)
                msg += "Requirements (should be all 0s):\n%s" % str(req1)
                warnings.warn(Warning(msg))

        # calculate the contributions from the eigenvalues
        gevalsA = grad_evals.unsqueeze(-2) * evecs  # (*BAM, na, neig)

        # calculate the contributions from the eigenvectors
        with M.uselinopparams(
                *mparams) if M is not None else dummy_context_manager():
            # orthogonalize the grad_evecs with evecs
            B = ortho(grad_evecs, evecs, D=idx_degen, M=M, mright=False)

            # Based on test cases, complex datatype is more likely to suffer from
            # singularity error when doing the inverse. Therefore, I add a small
            # offset here to prevent that from happening
            if torch.is_complex(B):
                evals_offset = evals + 1e-14
            else:
                evals_offset = evals

            with A.uselinopparams(*params):
                gevecs = solve(A,
                               -B,
                               evals_offset,
                               M,
                               bck_options=ctx.bck_config,
                               **ctx.bck_config)  # (*BAM, na, neig)

            # orthogonalize gevecs w.r.t. evecs
            gevecsA = ortho(gevecs, evecs, D=None, M=M, mright=True)

        # accummulate the gradient contributions
        gaccumA = gevalsA + gevecsA
        grad_params = torch.autograd.grad(
            outputs=(loss,),
            inputs=params,
            grad_outputs=(gaccumA,),
            create_graph=torch.is_grad_enabled(),
        )

        grad_mparams = []
        if ctx.M is not None:
            with torch.enable_grad():
                mparams = [p.clone().requires_grad_() for p in mparams]
                with M.uselinopparams(*mparams):
                    mloss = M.mm(evecs)  # (*BAM, na, neig)
            gevalsM = -gevalsA * evals.unsqueeze(-2)
            gevecsM = -gevecsA * evals.unsqueeze(-2)

            # the contribution from the parallel elements
            gevecsM_par = (-0.5 * torch.einsum(
                "...ae,...ae->...e", grad_evecs,
                evecs.conj())).unsqueeze(-2) * evecs  # (*BAM, na, neig)

            gaccumM = gevalsM + gevecsM + gevecsM_par
            grad_mparams = torch.autograd.grad(
                outputs=(mloss,),
                inputs=mparams,
                grad_outputs=(gaccumM,),
                create_graph=torch.is_grad_enabled(),
            )

        return (None, None, None, None, None, None, None, *grad_params,
                *grad_mparams)


def _check_degen(evals: torch.Tensor, degen_atol: float, degen_rtol: float) -> \
        Tuple[torch.Tensor, bool]:
    """Check the degeneracy of the eigenvalues

    Examples
    --------
    >>> import torch
    >>> evals = torch.tensor([1, 1, 2, 3, 3, 3, 4, 5, 5])
    >>> degen_atol = 0.1
    >>> degen_rtol = 0.1
    >>> idx_degen, isdegenerate = _check_degen(evals, degen_atol, degen_rtol)
    >>> idx_degen.shape
    torch.Size([9, 9])
    >>> isdegenerate
    True

    Parameters
    ----------
    evals: torch.Tensor
        Eigenvalues of the linear operator. Shape: ``(*BAM, neig)``
    degen_atol: float
        Minimum absolute difference between two eigenvalues to be treated as degenerate.
    degen_rtol: float
        Minimum relative difference between two eigenvalues to be treated as degenerate.

    Returns
    -------
    idx_degen: torch.Tensor
        The degeneracy map. Shape: ``(*BAM, neig, neig)``
    isdegenerate: bool
        Whether the eigenvalues are degenerate

    """
    # evals: (*BAM, neig)

    # get the index of degeneracies
    evals_diff = torch.abs(evals.unsqueeze(-2) -
                           evals.unsqueeze(-1))  # (*BAM, neig, neig)
    degen_thrsh = degen_atol + degen_rtol * torch.abs(evals).unsqueeze(-1)
    idx_degen = (evals_diff < degen_thrsh).to(evals.dtype)
    isdegenerate = bool(torch.sum(idx_degen) > torch.numel(evals))
    return idx_degen, isdegenerate


def ortho(A: torch.Tensor,
          B: torch.Tensor,
          *,
          D: Optional[torch.Tensor] = None,
          M: Optional[LinearOperator] = None,
          mright: bool = False) -> torch.Tensor:
    """Orthogonalize A w.r.t. B

    Examples
    --------
    >>> import torch
    >>> A = torch.tensor([[1, 2], [3, 4]])
    >>> B = torch.tensor([[1, 0], [0, 1]])
    >>> ortho(A, B)
    tensor([[0, 2],
            [3, 0]])

    Parameters
    ----------
    A: torch.Tensor
        The tensor to be orthogonalized. Shape: ``(*BAM, na, neig)``
    B: torch.Tensor
        The tensor to be orthogonalized against. Shape: ``(*BAM, na, neig)``
    D: torch.Tensor or None
        The degeneracy map. If None, it is identity matrix. Shape: ``(*BAM, neig, neig)``
    M: LinearOperator or None
        The overlap matrix. If None, identity matrix is used. Shape: ``(*BM, q, q)``
    mright: bool
        Whether to operate M at the right or at the left

    Returns
    -------
    torch.Tensor
        The orthogonalized tensor. Shape: ``(*BAM, na, neig)``

    """
    if D is None:
        # contracted using opt_einsum
        str1 = "...rc,...rc->...c"
        Bconj = B.conj()
        if M is None:
            return A - torch.einsum(str1, A, Bconj).unsqueeze(-2) * B
        elif mright:
            return A - torch.einsum(str1, M.mm(A), Bconj).unsqueeze(-2) * B
        else:
            return A - M.mm(torch.einsum(str1, A, Bconj).unsqueeze(-2) * B)
    else:
        BH = B.transpose(-2, -1).conj()
        if M is None:
            DBHA = D * torch.matmul(BH, A)
            return A - torch.matmul(B, DBHA)
        elif mright:
            DBHA = D * torch.matmul(BH, M.mm(A))
            return A - torch.matmul(B, DBHA)
        else:
            DBHA = D * torch.matmul(BH, A)
            return A - M.mm(torch.matmul(B, DBHA))


def exacteig(A: LinearOperator, neig: int, mode: str,
             M: Optional[LinearOperator]) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Eigendecomposition using explicit matrix construction.
    No additional option for this method.

    Examples
    --------
    >>> import torch
    >>> import numpy as np
    >>> from deepchem.utils.differentiation_utils import LinearOperator
    >>> A = LinearOperator.m(torch.rand(2, 2))
    >>> neig = 2
    >>> mode = "lowest"
    >>> M = None
    >>> evals, evecs = exacteig(A, neig, mode, M)
    >>> evals.shape
    torch.Size([2])
    >>> evecs.shape
    torch.Size([2, 2])

    Parameters
    ----------
    A: LinearOperator
        Linear operator to be diagonalized. Shape: ``(*BA, q, q)``.
    neig: int
        Number of eigenvalues and eigenvectors to be calculated.
    mode: str
        Mode of the eigenvalues to be calculated (``"lowest"``, ``"uppest"``)
    M: Optional[LinearOperator] (default None)
        The overlap matrix. If None, identity matrix is used. Shape: ``(*BM, q, q)``.

    Returns
    -------
    evals: torch.Tensor
        Eigenvalues of the linear operator.
    evecs: torch.Tensor
        Eigenvectors of the linear operator.

    Warnings
    --------
    * As this method construct the linear operators explicitly, it might requires
      a large memory.

    """
    Amatrix = A.fullmatrix()  # (*BA, q, q)
    if M is None:
        # evals, evecs = torch.linalg.eigh(Amatrix, eigenvectors=True)  # (*BA, q), (*BA, q, q)
        evals, evecs = degen_symeig.apply(Amatrix)  # (*BA, q, q)
        return _take_eigpairs(evals, evecs, neig, mode)
    else:
        Mmatrix = M.fullmatrix()  # (*BM, q, q)

        # M decomposition to make A symmetric
        # it is done this way to make it numerically stable in avoiding
        # complex eigenvalues for (near-)degenerate case
        L = torch.linalg.cholesky(Mmatrix)  # (*BM, q, q)
        Linv = torch.inverse(L)  # (*BM, q, q)
        LinvT = Linv.transpose(-2, -1).conj()  # (*BM, q, q)
        A2 = torch.matmul(Linv, torch.matmul(Amatrix, LinvT))  # (*BAM, q, q)

        # calculate the eigenvalues and eigenvectors
        # (the eigvecs are normalized in M-space)
        # evals, evecs = torch.linalg.eigh(A2, eigenvectors=True)  # (*BAM, q, q)
        evals, evecs = degen_symeig.apply(A2)  # (*BAM, q, q)
        evals, evecs = _take_eigpairs(evals, evecs, neig,
                                      mode)  # (*BAM, neig) and (*BAM, q, neig)
        evecs = torch.matmul(LinvT, evecs)
        return evals, evecs


# temporary solution to https://github.com/pytorch/pytorch/issues/47599
class degen_symeig(torch.autograd.Function):
    """A wrapper for torch.linalg.eigh to avoid complex eigenvalues for degenerate case.

    Examples
    --------
    >>> import torch
    >>> import numpy as np
    >>> from deepchem.utils.differentiation_utils import LinearOperator
    >>> A = LinearOperator.m(torch.rand(2, 2))
    >>> evals, evecs = degen_symeig.apply(A.fullmatrix())
    >>> evals.shape
    torch.Size([2])
    >>> evecs.shape
    torch.Size([2, 2])

    """

    @staticmethod
    def forward(ctx, A):
        """Calculate the eigenvalues and eigenvectors of a symmetric matrix.

        Parameters
        ----------
        A: torch.Tensor
            The symmetric matrix to be diagonalized. Shape: ``(*BA, q, q)``.

        Returns
        -------
        eival: torch.Tensor
            Eigenvalues of the linear operator.
        eivec: torch.Tensor
            Eigenvectors of the linear operator.

        """
        eival, eivec = torch.linalg.eigh(A)
        ctx.save_for_backward(eival, eivec)
        return eival, eivec

    @staticmethod
    def backward(ctx, grad_eival, grad_eivec):
        """Calculate the gradient of the eigenvalues and eigenvectors of a symmetric matrix.

        Parameters
        ----------
        grad_eival: torch.Tensor
            The gradient of the eigenvalues. Shape: ``(*BA, q)``.
        grad_eivec: torch.Tensor
            The gradient of the eigenvectors. Shape: ``(*BA, q, q)``.

        Returns
        -------
        result: torch.Tensor
            The gradient of the symmetric matrix. Shape: ``(*BA, q, q)``.

        """
        eival, eivec = ctx.saved_tensors
        min_threshold = torch.finfo(eival.dtype).eps**0.6
        eivect = eivec.transpose(-2, -1).conj()

        # remove the degenerate part
        # see https://arxiv.org/pdf/2011.04366.pdf
        if grad_eivec is not None:
            # take the contribution from the eivec
            F = eival.unsqueeze(-2) - eival.unsqueeze(-1)
            idx = torch.abs(F) <= min_threshold
            F[idx] = float("inf")

            F = F.pow(-1)
            F = F * torch.matmul(eivect, grad_eivec)
            result = torch.matmul(eivec, torch.matmul(F, eivect))
        else:
            result = torch.zeros_like(eivec)

        # calculate the contribution from the eival
        if grad_eival is not None:
            result += torch.matmul(eivec, grad_eival.unsqueeze(-1) * eivect)

        # symmetrize to reduce numerical instability
        result = (result + result.transpose(-2, -1).conj()) * 0.5
        return result


def davidson(A: LinearOperator,
             neig: int,
             mode: str,
             M: Optional[LinearOperator] = None,
             max_niter: int = 1000,
             nguess: Optional[int] = None,
             v_init: str = "randn",
             max_addition: Optional[int] = None,
             min_eps: float = 1e-6,
             verbose: bool = False,
             **unused) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Using Davidson method for large sparse matrix eigendecomposition [2]_.

    Examples
    --------
    >>> import torch
    >>> import numpy as np
    >>> from deepchem.utils.differentiation_utils import LinearOperator
    >>> A = LinearOperator.m(torch.rand(2, 2))
    >>> neig = 2
    >>> mode = "lowest"
    >>> eigen_val, eigen_vec = davidson(A, neig, mode)

    Parameters
    ----------
    A: LinearOperator
        Linear operator to be diagonalized. Shape: ``(*BA, q, q)``.
    neig: int
        Number of eigenvalues and eigenvectors to be calculated.
    mode: str
        Mode of the eigenvalues to be calculated (``"lowest"``, ``"uppest"``)
    M: Optional[LinearOperator] (default None)
        The overlap matrix. If None, identity matrix is used. Shape: ``(*BM, q, q)``.
    max_niter: int
        Maximum number of iterations
    v_init: str
        Mode of the initial guess (``"randn"``, ``"rand"``, ``"eye"``)
    max_addition: int or None
        Maximum number of new guesses to be added to the collected vectors.
        If None, set to ``neig``.
    min_eps: float
        Minimum residual error to be stopped
    verbose: bool
        Option to be verbose

    Returns
    -------
    evals: torch.Tensor
        Eigenvalues of the linear operator.
    evecs: torch.Tensor
        Eigenvectors of the linear operator.

    References
    ----------
    .. [2] P. Arbenz, "Lecture Notes on Solving Large Scale Eigenvalue Problems"
           http://people.inf.ethz.ch/arbenz/ewp/Lnotes/chapter12.pdf

    """
    # TODO: optimize for large linear operator and strict min_eps
    # Ideas:
    # (1) use better strategy to get the estimate on eigenvalues
    # (2) use restart strategy

    if nguess is None:
        nguess = neig
    if max_addition is None:
        max_addition = neig

    # get the shape of the transformation
    na = A.shape[-1]
    if M is None:
        bcast_dims = A.shape[:-2]
    else:
        bcast_dims = get_bcasted_dims(A.shape[:-2], M.shape[:-2])
    dtype = A.dtype
    device = A.device

    prev_eigvalT = None

    # set up the initial guess
    V = _set_initial_v(v_init.lower(),
                       dtype,
                       device,
                       bcast_dims,
                       na,
                       nguess,
                       M=M)  # (*BAM, na, nguess)

    best_resid: Union[float, torch.Tensor] = float("inf")
    AV = A.mm(V)
    for i in range(max_niter):
        VT = V.transpose(-2, -1)  # (*BAM,nguess,na)
        # Can be optimized by saving AV from the previous iteration and only
        # operate AV for the new V. This works because the old V has already
        # been orthogonalized, so it will stay the same
        # AV = A.mm(V) # (*BAM,na,nguess)
        T = torch.matmul(VT, AV)  # (*BAM,nguess,nguess)

        # eigvals are sorted from the lowest
        # eval: (*BAM, nguess), evec: (*BAM, nguess, nguess)
        eigvalT, eigvecT = torch.linalg.eigh(T)
        eigvalT, eigvecT = _take_eigpairs(
            eigvalT, eigvecT, neig,
            mode)  # (*BAM, neig) and (*BAM, nguess, neig)

        # calculate the eigenvectors of A
        eigvecA = torch.matmul(V, eigvecT)  # (*BAM, na, neig)

        # calculate the residual
        AVs = torch.matmul(AV, eigvecT)  # (*BAM, na, neig)
        LVs = eigvalT.unsqueeze(-2) * eigvecA  # (*BAM, na, neig)
        if M is not None:
            LVs = M.mm(LVs)
        resid = AVs - LVs  # (*BAM, na, neig)

        # print information and check convergence
        max_resid = resid.abs().max()
        if prev_eigvalT is not None:
            deigval = eigvalT - prev_eigvalT
            max_deigval = deigval.abs().max()
            if verbose:
                print("Iter %3d (guess size: %d): resid: %.3e, devals: %.3e" %
                      (i + 1, nguess, max_resid, max_deigval))  # type:ignore

        if max_resid < best_resid:
            best_resid = max_resid
            best_eigvals = eigvalT
            best_eigvecs = eigvecA
        if max_resid < min_eps:
            break
        if AV.shape[-1] == AV.shape[-2]:
            break
        prev_eigvalT = eigvalT

        # apply the preconditioner
        t = -resid  # (*BAM, na, neig)

        # orthogonalize t with the rest of the V
        t = to_fortran_order(t)
        Vnew = torch.cat((V, t), dim=-1)
        if Vnew.shape[-1] > Vnew.shape[-2]:
            Vnew = Vnew[..., :Vnew.shape[-2]]
        nadd = Vnew.shape[-1] - V.shape[-1]
        nguess = nguess + nadd
        if M is not None:
            MV_ = M.mm(Vnew)
            V, R = tallqr(Vnew, MV=MV_)
        else:
            V, R = tallqr(Vnew)
        AVnew = A.mm(V[..., -nadd:])  # (*BAM,na,nadd)
        AVnew = to_fortran_order(AVnew)
        AV = torch.cat((AV, AVnew), dim=-1)

    eigvals = best_eigvals  # (*BAM, neig)
    eigvecs = best_eigvecs  # (*BAM, na, neig)
    return eigvals, eigvecs


def _set_initial_v(vinit_type: str,
                   dtype: torch.dtype,
                   device: torch.device,
                   batch_dims: Sequence,
                   na: int,
                   nguess: int,
                   M: Optional[LinearOperator] = None) -> torch.Tensor:
    """Set the initial guess for the eigenvectors.

    Examples
    --------
    >>> import torch
    >>> vinit_type = "eye"
    >>> dtype = torch.float64
    >>> device = torch.device("cpu")
    >>> batch_dims = (2, 3)
    >>> na = 4
    >>> nguess = 2
    >>> M = None
    >>> V = _set_initial_v(vinit_type, dtype, device, batch_dims, na, nguess, M)
    >>> V
    tensor([[[[1., 0.],
              [0., 1.],
              [0., 0.],
              [0., 0.]],
    <BLANKLINE>
             [[1., 0.],
              [0., 1.],
              [0., 0.],
              [0., 0.]],
    <BLANKLINE>
             [[1., 0.],
              [0., 1.],
              [0., 0.],
              [0., 0.]]],
    <BLANKLINE>
    <BLANKLINE>
            [[[1., 0.],
              [0., 1.],
              [0., 0.],
              [0., 0.]],
    <BLANKLINE>
             [[1., 0.],
              [0., 1.],
              [0., 0.],
              [0., 0.]],
    <BLANKLINE>
             [[1., 0.],
              [0., 1.],
              [0., 0.],
              [0., 0.]]]], dtype=torch.float64)

    Parameters
    ----------
    vinit_type: str
        Mode of the initial guess (``"randn"``, ``"rand"``, ``"eye"``)
    dtype: torch.dtype
        Data type of the initial guess.
    device: torch.device
        Device of the initial guess.
    batch_dims: Sequence
        Batch dimensions of the initial guess.
    na: int
        Number of basis functions.
    nguess: int
        Number of initial guesses.
    M: Optional[LinearOperator] (default None)
        The overlap matrix. If None, identity matrix is used.

    Returns
    -------
    V: torch.Tensor
        Initial guess for the eigenvectors.

    """

    torch.manual_seed(12421)
    if vinit_type == "eye":
        nbatch = functools.reduce(lambda x, y: x * y, batch_dims, 1)
        V = torch.eye(na, nguess, dtype=dtype,
                      device=device).unsqueeze(0).repeat(nbatch, 1, 1).reshape(
                          *batch_dims, na, nguess)
    elif vinit_type == "randn":
        V = torch.randn((*batch_dims, na, nguess), dtype=dtype, device=device)
    elif vinit_type == "random" or vinit_type == "rand":
        V = torch.rand((*batch_dims, na, nguess), dtype=dtype, device=device)
    else:
        raise ValueError("Unknown v_init type: %s" % vinit_type)

    # orthogonalize V
    if isinstance(M, LinearOperator):
        V, R = tallqr(V, MV=M.mm(V))
    else:
        V, R = tallqr(V)
    return V


def _take_eigpairs(eival: torch.Tensor, eivec: torch.Tensor, neig: int,
                   mode: str):
    """Take the eigenpairs from the eigendecomposition.

    Examples
    --------
    >>> import torch
    >>> eival = torch.tensor([[1., 2., 3.], [4., 5., 6.]])
    >>> eivec = torch.tensor([[[1., 2., 3.], [4., 5., 6.], [7., 8., 9.]],
    ...                       [[1., 2., 3.], [4., 5., 6.], [7., 8., 9.]]])
    >>> neig = 2
    >>> mode = "lowest"
    >>> eival, eivec = _take_eigpairs(eival, eivec, neig, mode)
    >>> eival
    tensor([[1., 2.],
            [4., 5.]])
    >>> eivec
    tensor([[[1., 2.],
             [4., 5.],
             [7., 8.]],
    <BLANKLINE>
            [[1., 2.],
             [4., 5.],
             [7., 8.]]])

    Parameters
    ----------
    eival: torch.Tensor
        Eigenvalues of the linear operator. Shape: ``(*BV, na)``.
    eivec: torch.Tensor
        Eigenvectors of the linear operator. Shape: ``(*BV, na, na)``.
    neig: int
        Number of eigenvalues and eigenvectors to be calculated.
    mode: str
        Mode of the eigenvalues to be calculated (``"lowest"``, ``"uppest"``)

    Returns
    -------
    eival: torch.Tensor
        Eigenvalues of the linear operator.
    eivec: torch.Tensor
        Eigenvectors of the linear operator.

    """
    if mode == "lowest":
        eival = eival[..., :neig]
        eivec = eivec[..., :neig]
    else:
        eival = eival[..., -neig:]
        eivec = eivec[..., -neig:]
    return eival, eivec
